import copy
import typing

from . import enums
from . import tag as _tag
from .bundle import LilyPondFormatBundle
from .fsv import format_scheme_value
from .lyenv import contexts, grob_interfaces
from .storage import FormatSpecification, StorageFormatManager
from .string import String


def format_embedded_scheme_value(value):
    result = format_scheme_value(value)
    if isinstance(value, bool):
        result = "#" + result
    return result


def format_lilypond_attribute(attribute) -> str:
    assert isinstance(attribute, str), repr(attribute)
    attribute = attribute.replace("__", ".")
    result = attribute.replace("_", "-")
    return result


lilypond_color_constants = (
    "black",
    "blue",
    "center",
    "cyan",
    "darkblue",
    "darkcyan",
    "darkgreen",
    "darkmagenta",
    "darkred",
    "darkyellow",
    "down",
    "green",
    "grey",
    "left",
    "magenta",
    "red",
    "right",
    "up",
    "white",
    "yellow",
)


def format_lilypond_value(argument) -> str:
    if "_get_lilypond_format" in dir(argument):
        return argument._get_lilypond_format()
    if argument is True:
        return "##t"
    if argument is False:
        return "##f"
    if argument in (
        enums.Up,
        enums.Down,
        enums.Left,
        enums.Right,
        enums.Center,
    ):
        return rf"#{repr(argument).lower()}"
    if argument in lilypond_color_constants:
        return rf"#{argument}"
    if isinstance(argument, tuple) and len(argument) == 2:
        return f"#'({argument[0]} . {argument[1]})"
    return str(argument)


def make_lilypond_override_string(
    grob, attribute, value, context=None, once=False
) -> str:
    # camel_name = String(grob).to_upper_camel_case()
    # assert grob == camel_name, repr((grob, camel_name))
    grob = String(grob).to_upper_camel_case()
    attribute = format_lilypond_attribute(attribute)
    value = format_lilypond_value(value)
    if context is not None:
        context = String(context).capitalize_start() + "."
    else:
        context = ""
    if once is True:
        once = r"\once "
    else:
        once = ""
    result = rf"{once}\override {context}{grob}.{attribute} = {value}"
    return result


def make_lilypond_revert_string(grob, attribute, context=None) -> str:
    # camel_name = String(grob).to_upper_camel_case()
    # assert grob == camel_name, repr((grob, camel_name))
    grob = String(grob).to_upper_camel_case()
    dotted = format_lilypond_attribute(attribute)
    if context is not None:
        # camel_name = String(context).to_upper_camel_case()
        # assert context == camel_name, repr((context, camel_name))
        context = String(context).to_upper_camel_case()
        context += "."
    else:
        context = ""
    result = rf"\revert {context}{grob}.{dotted}"
    return result


def make_lilypond_tweak_string(
    attribute, value, *, directed=True, grob=None, literal=None
) -> str:
    if grob is not None:
        # camel_name = String(grob).to_upper_camel_case()
        # assert grob == camel_name, repr((grob, camel_name))
        grob = String(grob).to_upper_camel_case()
        grob += "."
    else:
        grob = ""
    attribute = format_lilypond_attribute(attribute)
    if not literal:
        value = format_lilypond_value(value)
    string = rf"\tweak {grob}{attribute} {value}"
    if directed:
        string = "- " + string
    return string


class LilyPondLiteral:
    r"""
    LilyPond literal.

    ..  container:: example

        Dotted slur:

        >>> staff = abjad.Staff("c'8 d'8 e'8 f'8")
        >>> abjad.slur(staff[:])
        >>> literal = abjad.LilyPondLiteral(r"\slurDotted")
        >>> abjad.attach(literal, staff[0])
        >>> abjad.show(staff) # doctest: +SKIP

        ..  docs::

            >>> string = abjad.lilypond(staff)
            >>> print(string)
            \new Staff
            {
                \slurDotted
                c'8
                (
                d'8
                e'8
                f'8
                )
            }

    ..  container:: example

        Use the absolute before and absolute after format slots like this:

        >>> staff = abjad.Staff("c'8 d'8 e'8 f'8")
        >>> abjad.slur(staff[:])
        >>> literal = abjad.LilyPondLiteral(r"\slurDotted")
        >>> abjad.attach(literal, staff[0])
        >>> literal = abjad.LilyPondLiteral("", format_slot="absolute_before")
        >>> abjad.attach(literal, staff[0])
        >>> literal = abjad.LilyPondLiteral(
        ...     "% before all formatting",
        ...     format_slot="absolute_before",
        ...     )
        >>> abjad.attach(literal, staff[0])
        >>> literal = abjad.LilyPondLiteral("", format_slot="absolute_after")
        >>> abjad.attach(literal, staff[-1])
        >>> abjad.show(staff) # doctest: +SKIP

        ..  docs::

            >>> string = abjad.lilypond(staff)
            >>> print(string)
            \new Staff
            {
            <BLANKLINE>
                % before all formatting
                \slurDotted
                c'8
                (
                d'8
                e'8
                f'8
                )
            <BLANKLINE>
            }

    ..  container:: example

        LilyPond literals can be tagged:

        >>> staff = abjad.Staff("c'8 d'8 e'8 f'8")
        >>> abjad.slur(staff[:])
        >>> literal = abjad.LilyPondLiteral(r"\slurDotted")
        >>> abjad.attach(literal, staff[0], tag=abjad.Tag("+PARTS"))
        >>> abjad.show(staff) # doctest: +SKIP

        >>> string = abjad.lilypond(staff, tags=True)
        >>> print(string)
        \new Staff
        {
            %! +PARTS
            \slurDotted
            c'8
            (
            d'8
            e'8
            f'8
            )
        }

    ..  container:: example

        Multiline input is allowed:

        >>> staff = abjad.Staff("c'8 d'8 e'8 f'8")
        >>> abjad.slur(staff[:])
        >>> lines = [
        ...     r"\stopStaff",
        ...     r"\startStaff",
        ...     r"\once \override Staff.StaffSymbol.color = #red",
        ...     ]
        >>> literal = abjad.LilyPondLiteral(lines)
        >>> abjad.attach(literal, staff[2], tag=abjad.Tag("+PARTS"))
        >>> abjad.show(staff) # doctest: +SKIP

        >>> string = abjad.lilypond(staff, tags=True)
        >>> print(string)
        \new Staff
        {
            c'8
            (
            d'8
            %! +PARTS
            \stopStaff
            %! +PARTS
            \startStaff
            %! +PARTS
            \once \override Staff.StaffSymbol.color = #red
            e'8
            f'8
            )
        }

    ..  container:: example

        REGRESSION. Duplicate literals are allowed:

        >>> staff = abjad.Staff("c'4 d' e' f'")
        >>> literal = abjad.LilyPondLiteral("% text")
        >>> abjad.attach(literal, staff[0])
        >>> literal = abjad.LilyPondLiteral("% text")
        >>> abjad.attach(literal, staff[0])

        >>> string = abjad.lilypond(staff)
        >>> print(string)
        \new Staff
        {
            % text
            % text
            c'4
            d'4
            e'4
            f'4
        }

    """

    ### CLASS VARIABLES ###

    __slots__ = ("_argument", "_directed", "_format_slot", "_tweaks")

    _allowable_format_slots = (
        "absolute_after",
        "absolute_before",
        "after",
        "before",
        "closing",
        "opening",
    )

    _can_attach_to_containers = True

    _format_leaf_children = False

    ### INITIALIZER ###

    def __init__(
        self,
        argument: typing.Union[str, typing.List[str]] = "",
        # TODO: probaby change default to "before"
        format_slot: str = "opening",
        *,
        directed: bool = None,
        tweaks: "TweakInterface" = None,
    ) -> None:
        self._argument = argument
        assert format_slot in self._allowable_format_slots, repr(format_slot)
        self._format_slot = format_slot
        if directed is not None:
            directed = bool(directed)
        self._directed = directed
        if tweaks is not None:
            assert isinstance(tweaks, TweakInterface), repr(tweaks)
        self._tweaks = TweakInterface.set_tweaks(self, tweaks)

    ### SPECIAL METHODS ###

    def __eq__(self, argument) -> bool:
        """
        Is true when all initialization values of Abjad value object equal
        the initialization values of ``argument``.
        """
        return StorageFormatManager.compare_objects(self, argument)

    def __hash__(self) -> int:
        """
        Hashes Abjad value object.
        """
        hash_values = StorageFormatManager(self).get_hash_values()
        try:
            result = hash(hash_values)
        except TypeError:
            raise TypeError(f"unhashable type: {self}")
        return result

    def __repr__(self) -> str:
        """
        Gets interpreter representation.
        """
        return StorageFormatManager(self).get_repr_format()

    ### PRIVATE METHODS ###

    # TODO: activate this:
    #    def _before_attach(self, component):
    #        if self.format_slot not in component._allowable_format_slots:
    #            message = f"{type(component).__name__} does not accept"
    #            message += f" format slot {repr(self.format_slot)}."
    #            raise Exception(message)

    def _get_format_pieces(self):
        if isinstance(self.argument, str):
            return [self.argument]
        assert isinstance(self.argument, list)
        return self.argument[:]

    def _get_format_specification(self):
        return FormatSpecification(
            client=self,
            storage_format_args_values=[self.argument],
            storage_format_is_indented=False,
        )

    def _get_lilypond_format_bundle(self, component=None):
        bundle = LilyPondFormatBundle()
        format_slot = bundle.get(self.format_slot)
        if self.tweaks:
            tweaks = self.tweaks._list_format_contributions(directed=self.directed)
            format_slot.commands.extend(tweaks)
        pieces = self._get_format_pieces()
        format_slot.commands.extend(pieces)
        return bundle

    ### PUBLIC PROPERTIES ###

    @property
    def argument(self) -> typing.Union[str, typing.List[str]]:
        r"""
        Gets argument of LilyPond literal.

        ..  container:: example

            >>> literal = abjad.LilyPondLiteral(r"\slurDotted")
            >>> literal.argument
            '\\slurDotted'

        """
        return self._argument

    @property
    def directed(self) -> typing.Optional[bool]:
        r"""
        Is true when literal is directed.

        ..  container:: example

            Directed literal:

            >>> staff = abjad.Staff("c'4 d' e' f'")
            >>> literal = abjad.LilyPondLiteral(r"\f", "after", directed=True)
            >>> abjad.tweak(literal).color = "#blue"
            >>> abjad.tweak(literal).DynamicLineSpanner.staff_padding = 5
            >>> abjad.attach(literal, staff[0])
            >>> abjad.show(staff) # doctest: +SKIP

            ..  docs::

                >>> string = abjad.lilypond(staff)
                >>> print(string)
                \new Staff
                {
                    c'4
                    - \tweak DynamicLineSpanner.staff-padding 5
                    - \tweak color #blue
                    \f
                    d'4
                    e'4
                    f'4
                }

        ..  container:: example

            Nondirected literal:

            >>> staff = abjad.Staff("c'4 d' e' f'")
            >>> literal = abjad.LilyPondLiteral(
            ...     r"\breathe",
            ...     "after",
            ...     directed=False,
            ...     )
            >>> abjad.tweak(literal).color = "#blue"
            >>> abjad.attach(literal, staff[0])
            >>> abjad.show(staff) # doctest: +SKIP

            ..  docs::

                >>> string = abjad.lilypond(staff)
                >>> print(string)
                \new Staff
                {
                    c'4
                    \tweak color #blue
                    \breathe
                    d'4
                    e'4
                    f'4
                }

        Proper use of the ``directed`` property entails searching the LilyPond
        docs to understand whether LilyPond treats any particular command as
        directed or not. Most LilyPond commands are directed. LilyPond insists
        that a few commands (include ``\breathe``, ``\key``, ``\mark``) must
        not be directed.
        """
        return self._directed

    @property
    def format_slot(self) -> str:
        r"""
        Gets format slot of LilyPond literal.

        ..  container:: example

            >>> literal = abjad.LilyPondLiteral(r"\slurDotted")
            >>> literal.format_slot
            'opening'

        """
        return self._format_slot

    @property
    def tweaks(self) -> typing.Optional["TweakInterface"]:
        r"""
        Gets tweaks.

        ..  container:: example

            >>> staff = abjad.Staff("c'4 d' e' f'")
            >>> literal = abjad.LilyPondLiteral(r"\f", "after", directed=True)
            >>> abjad.tweak(literal).color = "#blue"
            >>> abjad.attach(literal, staff[0])
            >>> abjad.show(staff) # doctest: +SKIP

            ..  docs::

                >>> string = abjad.lilypond(staff)
                >>> print(string)
                \new Staff
                {
                    c'4
                    - \tweak color #blue
                    \f
                    d'4
                    e'4
                    f'4
                }

        """
        return self._tweaks


class Interface:
    """
    LilyPond name manager.

    Base class from which grob, setting and tweak managers inherit.
    """

    ### SPECIAL METHODS ###

    def __eq__(self, argument) -> bool:
        """
        Is true when ``argument`` is a LilyPond name manager with attribute
        pairs equal to those of this LilyPond name manager.

        ..  container:: example

            >>> note_1 = abjad.Note("c'4")
            >>> abjad.setting(note_1).Voice.auto_beaming = False
            >>> abjad.setting(note_1).Staff.tupletFullLength = True

            >>> note_2 = abjad.Note("c'4")
            >>> abjad.setting(note_2).Voice.auto_beaming = False
            >>> abjad.setting(note_2).Staff.tupletFullLength = True

            >>> note_3 = abjad.Note("c'4")
            >>> abjad.setting(note_3).Voice.auto_beaming = True

            >>> setting_1 = abjad.setting(note_1)
            >>> setting_2 = abjad.setting(note_2)
            >>> setting_3 = abjad.setting(note_3)

            >>> setting_1 == setting_1
            True
            >>> setting_1 == setting_2
            True
            >>> setting_1 == setting_3
            False

            >>> setting_2 == setting_1
            True
            >>> setting_2 == setting_2
            True
            >>> setting_2 == setting_3
            False

            >>> setting_3 == setting_1
            False
            >>> setting_3 == setting_2
            False
            >>> setting_3 == setting_3
            True

        ..  container:: example

            >>> note_1 = abjad.Note("c'4")
            >>> abjad.override(note_1).NoteHead.color = "#red"
            >>> abjad.override(note_1).Stem.color = "#red"

            >>> note_2 = abjad.Note("c'4")
            >>> abjad.override(note_2).NoteHead.color = "#red"
            >>> abjad.override(note_2).Stem.color = "#red"

            >>> note_3 = abjad.Note("c'4")
            >>> abjad.override(note_3).NoteHead.color = "#red"

            >>> override_1 = abjad.override(note_1)
            >>> override_2 = abjad.override(note_2)
            >>> override_3 = abjad.override(note_3)

            >>> override_1 == override_1
            True
            >>> override_1 == override_2
            True
            >>> override_1 == override_3
            False

            >>> override_2 == override_1
            True
            >>> override_2 == override_2
            True
            >>> override_2 == override_3
            False

            >>> override_3 == override_1
            False
            >>> override_3 == override_2
            False
            >>> override_3 == override_3
            True

        """
        if isinstance(argument, type(self)):
            attribute_pairs_1 = self._get_attribute_pairs()
            attribute_pairs_2 = argument._get_attribute_pairs()
            return attribute_pairs_1 == attribute_pairs_2
        return False

    def __getstate__(self) -> dict:
        """
        Gets object state.
        """
        return copy.deepcopy(vars(self))

    def __hash__(self) -> int:
        """
        Hashes LilyPond name manager.

        Redefined in tandem with __eq__.
        """
        return super().__hash__()

    def __repr__(self) -> str:
        """
        Gets interpreter representation of LilyPond name manager.
        """
        body_string = ""
        pairs = self._get_attribute_pairs()
        pairs = [str(_) for _ in pairs]
        body_string = ", ".join(pairs)
        return f"{type(self).__name__}({body_string})"

    def __setstate__(self, state) -> None:
        """
        Sets object state.
        """
        for key, value in state.items():
            self.__dict__[key] = value

    ### PRIVATE METHODS ###

    def _get_attribute_pairs(self):
        return list(sorted(vars(self).items()))


class LilyPondOverride:
    r"""
    LilyPond grob override.

    ..  container:: example

        >>> override = abjad.LilyPondOverride(
        ...    lilypond_type="Staff",
        ...    grob_name="TextSpanner",
        ...    once=True,
        ...    property_path=(
        ...        "bound-details",
        ...        "left",
        ...        "text",
        ...        ),
        ...    value=abjad.Markup(r"\bold { over pressure }"),
        ...    )

        >>> print(override.override_string)
        \once \override Staff.TextSpanner.bound-details.left.text = \markup {
            \bold
                {
                    over
                    pressure
                }
            }

    """

    ### CLASS VARIABLES ###

    __slots__ = (
        "_grob_name",
        "_is_revert",
        "_lilypond_type",
        "_once",
        "_property_path",
        "_value",
    )

    _format_leaf_children = False

    ### INITIALIZER ###

    def __init__(
        self,
        lilypond_type: str = None,
        grob_name: str = "NoteHead",
        once: bool = None,
        is_revert: bool = None,
        property_path: typing.Union[str, typing.Iterable[str]] = "color",
        value: typing.Any = "#red",
    ) -> None:
        if lilypond_type is not None:
            lilypond_type = str(lilypond_type)
        self._lilypond_type = lilypond_type
        assert grob_name
        self._grob_name = str(grob_name)
        if once is not None:
            once = bool(once)
        self._once = once
        if is_revert is not None:
            is_revert = bool(is_revert)
        self._is_revert = is_revert
        if isinstance(property_path, str):
            property_path_: typing.Tuple[str, ...] = (property_path,)
        else:
            property_path_ = tuple(property_path)
        assert isinstance(property_path_, tuple), repr(property_path_)
        assert all(isinstance(_, str) for _ in property_path_)
        assert all(_ != "" for _ in property_path_)
        self._property_path = property_path_
        self._value = value

    ### SPECIAL METHODS ###

    def __eq__(self, argument) -> bool:
        """
        Is true when ``argument`` is a LilyPond grob override with equivalent
        keyword values.
        """
        return super().__eq__(argument)

    def __hash__(self) -> int:
        """
        Hashes LilyPond grob override.

        Redefined in tandem with __eq__.
        """
        return super().__hash__()

    ### PRIVATE METHODS ###

    def _get_lilypond_format_bundle(self, component=None):
        bundle = LilyPondFormatBundle()
        if not self.once:
            revert_format = "\n".join(self.revert_format_pieces)
            bundle.grob_reverts.append(revert_format)
        if not self.is_revert:
            override_format = "\n".join(self.override_format_pieces)
            bundle.grob_overrides.append(override_format)
        return bundle

    def _override_property_path_string(self):
        parts = []
        if self.lilypond_type is not None:
            parts.append(self.lilypond_type)
        parts.append(self.grob_name)
        parts.extend(self.property_path)
        path = ".".join(parts)
        return path

    def _revert_property_path_string(self):
        parts = []
        if self.lilypond_type is not None:
            parts.append(self.lilypond_type)
        parts.append(self.grob_name)
        parts.append(self.property_path[0])
        path = ".".join(parts)
        return path

    ### PUBLIC PROPERTIES ###

    @property
    def grob_name(self) -> str:
        r"""
        Gets grob name.

        ..  container:: example

            >>> override = abjad.LilyPondOverride(
            ...     grob_name="Glissando",
            ...     property_path="style",
            ...     value="#'zigzag",
            ...     )
            >>> override.grob_name
            'Glissando'

        """
        return self._grob_name

    @property
    def is_revert(self) -> typing.Optional[bool]:
        r"""
        Is true if grob override is a grob revert.

        ..  container:: example

            >>> override = abjad.LilyPondOverride(
            ...     grob_name="Glissando",
            ...     property_path="style",
            ...     value="#'zigzag",
            ...     )
            >>> bool(override.is_revert)
            False

            >>> override = abjad.LilyPondOverride(
            ...     grob_name="Glissando",
            ...     is_revert=True,
            ...     property_path="style",
            ...     )
            >>> bool(override.is_revert)
            True

        """
        return self._is_revert

    @property
    def lilypond_type(self) -> typing.Optional[str]:
        r"""
        Gets LilyPond type of context.

        ..  container:: example

            >>> override = abjad.LilyPondOverride(
            ...    lilypond_type="Staff",
            ...    grob_name="TextSpanner",
            ...    once=True,
            ...    property_path=(
            ...        "bound-details",
            ...        "left",
            ...        "text",
            ...        ),
            ...    value=abjad.Markup(r"\bold { over pressure }"),
            ...    )
            >>> override.lilypond_type
            'Staff'

            >>> override = abjad.LilyPondOverride(
            ...     grob_name="Glissando",
            ...     property_path="style",
            ...     value="#'zigzag",
            ...     )
            >>> override.lilypond_type is None
            True

        """
        return self._lilypond_type

    @property
    def once(self) -> typing.Optional[bool]:
        r"""
        Is true when grob override is to be applied only once.

        ..  container:: example

            >>> override = abjad.LilyPondOverride(
            ...    lilypond_type="Staff",
            ...    grob_name="TextSpanner",
            ...    once=True,
            ...    property_path=(
            ...        "bound-details",
            ...        "left",
            ...        "text",
            ...        ),
            ...    value=abjad.Markup(r"\bold { over pressure }"),
            ...    )
            >>> bool(override.once)
            True

            >>> override = abjad.LilyPondOverride(
            ...     grob_name="Glissando",
            ...     property_path="style",
            ...     value="#'zigzag",
            ...     )
            >>> bool(override.once)
            False

        """
        return self._once

    @property
    def override_format_pieces(self) -> typing.Tuple[str, ...]:
        r"""
        Gets LilyPond grob override \override format pieces.

        ..  container:: example

            >>> override = abjad.LilyPondOverride(
            ...    lilypond_type="Staff",
            ...    grob_name="TextSpanner",
            ...    once=True,
            ...    property_path=(
            ...        "bound-details",
            ...        "left",
            ...        "text",
            ...        ),
            ...    value=abjad.Markup(r"\bold { over pressure }"),
            ...    )
            >>> for line in override.override_format_pieces:
            ...     line
            ...
            '\\once \\override Staff.TextSpanner.bound-details.left.text = \\markup {'
            '    \\bold'
            '        {'
            '            over'
            '            pressure'
            '        }'
            '    }'

        """
        result = []
        if self.once:
            result.append(r"\once")
        result.append(r"\override")
        result.append(self._override_property_path_string())
        result.append("=")
        string = format_embedded_scheme_value(self.value)
        # string = str(self.value)
        value_pieces = string.split("\n")
        result.append(value_pieces[0])
        result[:] = [" ".join(result)]
        result.extend(value_pieces[1:])
        return tuple(result)

    @property
    def override_string(self) -> str:
        r"""
        Gets LilyPond grob override \override string.

        ..  container:: example

            >>> override = abjad.LilyPondOverride(
            ...     grob_name="Glissando",
            ...     property_path="style",
            ...     value="#'zigzag",
            ...     )
            >>> override.override_string
            "\\override Glissando.style = #'zigzag"

        """
        return "\n".join(self.override_format_pieces)

    @property
    def property_path(self) -> typing.Tuple[str, ...]:
        r"""
        Gets LilyPond grob override property path.

        ..  container:: example

            >>> override = abjad.LilyPondOverride(
            ...    lilypond_type="Staff",
            ...    grob_name="TextSpanner",
            ...    once=True,
            ...    property_path=(
            ...        "bound-details",
            ...        "left",
            ...        "text",
            ...        ),
            ...    value=abjad.Markup(r"\bold { over pressure }"),
            ...    )
            >>> override.property_path
            ('bound-details', 'left', 'text')

        """
        return self._property_path

    @property
    def revert_format_pieces(self) -> typing.Tuple[str, ...]:
        r"""
        Gets LilyPond grob override \revert format pieces.

        ..  container:: example

            >>> override = abjad.LilyPondOverride(
            ...     grob_name="Glissando",
            ...     property_path="style",
            ...     value="#'zigzag",
            ...     )
            >>> override.revert_format_pieces
            ('\\revert Glissando.style',)

        """
        result = rf"\revert {self._revert_property_path_string()}"
        return (result,)

    @property
    def revert_string(self) -> str:
        r"""
        Gets LilyPond grob override \revert string.

        ..  container:: example

            >>> override = abjad.LilyPondOverride(
            ...     grob_name="Glissando",
            ...     property_path="style",
            ...     value="#'zigzag",
            ...     )
            >>> override.revert_string
            '\\revert Glissando.style'

        """
        return "\n".join(self.revert_format_pieces)

    @property
    def value(self) -> typing.Any:
        r"""
        Gets value of LilyPond grob override.

        ..  container:: example

            >>> override = abjad.LilyPondOverride(
            ...    lilypond_type="Staff",
            ...    grob_name="TextSpanner",
            ...    once=True,
            ...    property_path=(
            ...        "bound-details",
            ...        "left",
            ...        "text",
            ...        ),
            ...    value=abjad.Markup(r"\markup \bold { over pressure }", literal=True),
            ...    )
            >>> override.value
            Markup(contents=['\\markup \\bold { over pressure }'], literal=True)

        """
        return self._value

    ### PUBLIC METHODS ###

    def tweak_string(self, directed=True, grob=False) -> str:
        r"""
        Gets LilyPond grob override \tweak string.

        ..  container:: example

            >>> override = abjad.LilyPondOverride(
            ...     grob_name="Glissando",
            ...     property_path="style",
            ...     value="#'zigzag",
            ...     )
            >>> override.tweak_string()
            "- \\tweak style #'zigzag"

        ..  container:: example

            >>> override = abjad.LilyPondOverride(
            ...     grob_name="RehearsalMark",
            ...     property_path="color",
            ...     value="#red",
            ...     )
            >>> override.tweak_string(directed=False)
            '\\tweak color #red'

        ..  container:: example

            LilyPond literals are allowed:

            >>> override = abjad.LilyPondOverride(
            ...     grob_name="TextSpann",
            ...     property_path=("bound-details", "left-broken", "text"),
            ...     value=abjad.LilyPondLiteral(r"\markup \upright pont."),
            ...     )
            >>> override.tweak_string(directed=False)
            '\\tweak bound-details.left-broken.text \\markup \\upright pont.'

        """
        if directed:
            result = [r"- \tweak"]
        else:
            result = [r"\tweak"]
        if grob:
            property_path = (self.grob_name,) + self.property_path
        else:
            property_path = self.property_path
        string = ".".join(property_path)
        result.append(string)
        if isinstance(self.value, LilyPondLiteral):
            assert isinstance(self.value.argument, str)
            string = self.value.argument
        else:
            string = format_embedded_scheme_value(self.value)
            # string = str(self.value)
        result.append(string)
        return " ".join(result)


class LilyPondSetting:
    r"""
    LilyPond context setting.

    ..  container:: example

        >>> context_setting = abjad.LilyPondSetting(
        ...    lilypond_type="Score",
        ...    context_property="autoBeaming",
        ...    value="##f",
        ...    )

        >>> print("\n".join(context_setting.format_pieces))
        \set Score.autoBeaming = ##f

    """

    ### CLASS VARIABLES ###

    __slots__ = ("_context_property", "_lilypond_type", "_is_unset", "_value")

    _format_leaf_children = False

    ### INITIALIZER ###

    def __init__(
        self,
        lilypond_type: str = None,
        context_property: str = "autoBeaming",
        is_unset: bool = False,
        value: typing.Any = False,
    ) -> None:
        if lilypond_type is not None:
            lilypond_type = str(lilypond_type)
        self._lilypond_type = lilypond_type
        assert isinstance(context_property, str) and context_property
        self._context_property = context_property
        if is_unset is not None:
            is_unset = bool(is_unset)
        self._is_unset = is_unset
        self._value = value

    ### SPECIAL METHODS ###

    def __eq__(self, argument) -> bool:
        """
        Is true when ``argument`` is a LilyPond context setting with
        equivalent keyword values.
        """
        return super().__eq__(argument)

    def __hash__(self) -> int:
        """
        Hashes LilyPond context setting.

        Redefined in tandem with __eq__.
        """
        return super().__hash__()

    ### PRIVATE METHODS ###

    def _get_lilypond_format_bundle(self, component=None):
        bundle = LilyPondFormatBundle()
        string = "\n".join(self.format_pieces)
        bundle.context_settings.append(string)
        return bundle

    ### PUBLIC PROPERTIES ###

    @property
    def context_property(self) -> str:
        """
        Gets LilyPond context property name.
        """
        return self._context_property

    @property
    def format_pieces(self) -> typing.Tuple[str, ...]:
        r"""
        Gets LilyPond context setting ``\set`` or ``\unset`` format pieces.
        """
        result = []
        if not self.is_unset:
            result.append(r"\set")
        else:
            result.append(r"\unset")
        if self.lilypond_type is not None:
            string = f"{self.lilypond_type}.{self.context_property}"
            result.append(string)
        else:
            result.append(self.context_property)
        result.append("=")
        string = format_embedded_scheme_value(self.value)
        # assert string == str(self.value), repr((str(self.value), string))
        value_pieces = string.split("\n")
        result.append(value_pieces[0])
        result[:] = [" ".join(result)]
        result.extend(value_pieces[1:])
        return tuple(result)

    @property
    def is_unset(self) -> typing.Optional[bool]:
        """
        Is true if context setting unsets its value.
        """
        return self._is_unset

    @property
    def lilypond_type(self) -> typing.Optional[str]:
        """
        Gets LilyPond type.
        """
        return self._lilypond_type

    @property
    def value(self) -> typing.Any:
        """
        Gets value of LilyPond context setting.
        """
        return self._value


class OverrideInterface(Interface):
    """
    LilyPond grob name manager.

    ..  container:: example

        OverrideInterface instances are created by the
        ``abjad.override()`` factory function:

        >>> note = abjad.Note("c'4")
        >>> abjad.override(note)
        OverrideInterface()

    """

    ### SPECIAL METHODS ###

    def __getattr__(self, name) -> typing.Union[Interface, "OverrideInterface"]:
        r"""
        Gets Interface (or OverrideInterface) keyed to ``name``.

        ..  container:: example

            Somewhat confusingly, getting a grob name returns a
            Interface:

            >>> staff = abjad.Staff("c'4 d' e' f'")
            >>> abjad.override(staff[0]).NoteHead
            Interface()

            While getting a context name returns an override interface:

            >>> staff = abjad.Staff("c'4 d' e' f'")
            >>> abjad.override(staff[0]).Staff
            OverrideInterface()

            Which can then be deferenced to get an interface:

            >>> staff = abjad.Staff("c'4 d' e' f'")
            >>> abjad.override(staff[0]).Staff.NoteHead
            Interface()

        Note that the dot-chained user syntax is unproblematic. But the class of each
        manager returned in the chain is likely to be surprising at first encounter.
        """
        camel_name = String(name).to_upper_camel_case()
        # assert name == camel_name, repr((name, camel_name))
        if name.startswith("_"):
            try:
                return vars(self)[name]
            except KeyError:
                type_name = type(self).__name__
                raise AttributeError("{type_name!r} object has no attribute: {name!r}.")
        elif camel_name in contexts:
            try:
                return vars(self)["_" + name]
            except KeyError:
                context = OverrideInterface()
                vars(self)["_" + name] = context
                return context
        elif camel_name in grob_interfaces:
            try:
                return vars(self)[name]
            except KeyError:
                vars(self)[name] = Interface()
                return vars(self)[name]
        else:
            try:
                return vars(self)[name]
            except KeyError:
                type_name = type(self).__name__
                raise AttributeError(
                    f"{type_name!r} object has no attribute: {name!r}."
                )

    def __setattr__(self, attribute, value) -> None:
        """
        Sets attribute ``attribute`` of grob name manager to ``value``.
        """
        object.__setattr__(self, attribute, value)

    ### PRIVATE METHODS ###

    def _get_attribute_tuples(self):
        result = []
        for name, value in vars(self).items():
            if type(value) is Interface:
                grob, grob_proxy = name, value
                pairs = iter(vars(grob_proxy).items())
                for attribute, value in pairs:
                    triple = (grob, attribute, value)
                    result.append(triple)
            else:
                context, context_proxy = name.strip("_"), value
                for grob, grob_proxy in vars(context_proxy).items():
                    pairs = iter(vars(grob_proxy).items())
                    for attribute, value in pairs:
                        quadruple = (context, grob, attribute, value)
                        result.append(quadruple)
        result.sort()
        return tuple(result)

    def _list_format_contributions(self, contribution_type, once=False):
        assert contribution_type in ("override", "revert")
        result = []
        for attribute_tuple in self._get_attribute_tuples():
            if len(attribute_tuple) == 3:
                context = None
                grob = attribute_tuple[0]
                attribute = attribute_tuple[1]
                value = attribute_tuple[2]
            elif len(attribute_tuple) == 4:
                context = attribute_tuple[0]
                grob = attribute_tuple[1]
                attribute = attribute_tuple[2]
                value = attribute_tuple[3]
            else:
                raise ValueError(f"invalid attribute tuple: {attribute_tuple!r}.")
            if contribution_type == "override":
                override_string = make_lilypond_override_string(
                    grob, attribute, value, context=context, once=once
                )
                result.append(override_string)
            else:
                revert_string = make_lilypond_revert_string(
                    grob, attribute, context=context
                )
                result.append(revert_string)
        result.sort()
        return result


### FUNCTIONS ###


def override(argument):
    r"""
    Makes LilyPond grob name manager.

    ..  container:: example

        Overrides staff symbol color:

        >>> staff = abjad.Staff("c'4 e'4 d'4 f'4")
        >>> abjad.override(staff).StaffSymbol.color = "#red"
        >>> abjad.show(staff) # doctest: +SKIP

        ..  docs::

            >>> string = abjad.lilypond(staff)
            >>> print(string)
            \new Staff
            \with
            {
                \override StaffSymbol.color = #red
            }
            {
                c'4
                e'4
                d'4
                f'4
            }

    ..  container:: example

        Specify grob context like this:

        >>> staff = abjad.Staff("c'4 e'4 d'4 f'4")
        >>> abjad.override(staff[0]).Staff.StaffSymbol.color = "#blue"
        >>> abjad.show(staff) # doctest: +SKIP

        ..  docs::

            >>> string = abjad.lilypond(staff)
            >>> print(string)
            \new Staff
            {
                \once \override Staff.StaffSymbol.color = #blue
                c'4
                e'4
                d'4
                f'4
            }

    ..  container:: example

        Returns LilyPond grob name manager:

        >>> staff = abjad.Staff("c'4 e' d' f'")
        >>> abjad.override(staff)
        OverrideInterface()

    """
    if getattr(argument, "_overrides", None) is None:
        argument._overrides = OverrideInterface()
    return argument._overrides


class SettingInterface(Interface):
    """
    LilyPond setting name manager.

    ..  container:: example

        >>> note = abjad.Note("c'4")
        >>> abjad.setting(note)
        SettingInterface()

    """

    ### SPECIAL METHODS ###

    def __getattr__(self, name: str) -> typing.Any:
        r"""
        Gets arbitrary object keyed to ``name``.

        ..  container:: example

            >>> staff = abjad.Staff("c'4 d' e' f'")
            >>> abjad.setting(staff).instrumentName = abjad.Markup("Vn. I")
            >>> abjad.show(staff) # doctest: +SKIP

            ..  docs::

                >>> string = abjad.lilypond(staff)
                >>> print(string)
                \new Staff
                \with
                {
                    instrumentName = \markup { Vn. I }
                }
                {
                    c'4
                    d'4
                    e'4
                    f'4
                }

        ..  container:: example

            Returns arbitrary object keyed to ``name``:

            >>> abjad.setting(staff).instrumentName
            Markup(contents=['Vn. I'])

        """
        camel_name = String(name).to_upper_camel_case()
        # assert name == camel_name, repr((name, camel_name))
        if name.startswith("_"):
            try:
                return vars(self)[name]
            except KeyError:
                message = "{type(self).__name__!r} object has no attribute: {name!r}."
                raise AttributeError(message)
        elif camel_name in contexts:
            try:
                return vars(self)["_" + name]
            except KeyError:
                context = Interface()
                vars(self)["_" + name] = context
                return context
        else:
            try:
                return vars(self)[name]
            except KeyError:
                message = "{type(self).__name__!r} object has no attribute: {name!r}."
                raise AttributeError(message)

    ### PRIVATE METHODS ###

    def _format_in_with_block(self) -> typing.List[str]:
        strings = []
        for key, value in vars(self).items():
            assert isinstance(key, str), repr(key)
            name = key.split("_")
            first = name[0:1]
            rest = name[1:]
            rest = [x.title() for x in rest]
            name = first + rest
            string = "".join(name)
            # assert key == string, repr((key, string))
            value = format_lilypond_value(value)
            value_parts = value.split("\n")
            result = rf"{string!s} = {value_parts[0]!s}"
            pieces = [result]
            for part in value_parts[1:]:
                pieces.append(LilyPondFormatBundle.indent + part)
            string = "\n".join(pieces)
            strings.append(string)
        return strings

    def _format_inline(self, context=None) -> typing.List[str]:
        result = []
        for name, value in vars(self).items():
            # if we've found a leaf context namespace
            if name.startswith("_"):
                for x, y in vars(value).items():
                    if not x.startswith("_"):
                        string = self._format_inline_helper(x, y, name)
                        result.append(string)
            # otherwise we've found a default leaf context setting
            else:
                # parse default context setting
                string = self._format_inline_helper(name, value)
                result.append(string)
        return result

    def _format_inline_helper(self, name, value, context=None):
        name = name.split("_")
        first = name[0:1]
        rest = name[1:]
        rest = [x.title() for x in rest]
        name = first + rest
        name = "".join(name)
        value = format_lilypond_value(value)
        if context is not None:
            context_string = context[1:]
            context_string = context_string.split("_")
            context_string = [x.title() for x in context_string]
            context_string = "".join(context_string)
            context_string += "."
        else:
            context_string = ""
        result = rf"\set {context_string}{name} = {value}"
        return result

    def _get_attribute_tuples(self):
        result = []
        for name, value in vars(self).items():
            if type(value) is Interface:
                prefixed_context_name = name
                lilypond_type = prefixed_context_name.strip("_")
                context_proxy = value
                attribute_pairs = context_proxy._get_attribute_pairs()
                for attribute_name, attribute_value in attribute_pairs:
                    triple = (lilypond_type, attribute_name, attribute_value)
                    result.append(triple)
            else:
                attribute_name, attribute_value = name, value
                result.append((attribute_name, attribute_value))
        result.sort()
        return result


### FUNCTIONS ###


def setting(argument):
    r"""
    Makes LilyPond setting name manager.

    ..  container:: example

        Sets instrument name:

        >>> staff = abjad.Staff("c'4 e'4 d'4 f'4")
        >>> abjad.setting(staff).instrumentName = abjad.Markup("Vn. I")
        >>> abjad.show(staff) # doctest: +SKIP


        ..  docs::

            >>> string = abjad.lilypond(staff)
            >>> print(string)
            \new Staff
            \with
            {
                instrumentName = \markup { Vn. I }
            }
            {
                c'4
                e'4
                d'4
                f'4
            }

    ..  container:: example

        Returns LilyPond setting name manager:

        >>> abjad.setting(staff)
        SettingInterface(('instrumentName', Markup(contents=['Vn. I'])))

    """
    if getattr(argument, "_lilypond_setting_name_manager", None) is None:
        argument._lilypond_setting_name_manager = SettingInterface()
    return argument._lilypond_setting_name_manager


class TweakInterface(Interface):
    """
    LilyPond tweak manager.

    ..  container:: example

        Tweak managers are created by the ``abjad.tweak()`` factory function:

        >>> markup = abjad.Markup("Allegro", direction=abjad.Up)
        >>> abjad.tweak(markup)
        TweakInterface(('_literal', None))

        Set an attribute like this:

        >>> abjad.tweak(markup).color = "#red"

        The state of the tweak manager has changed:

        >>> abjad.tweak(markup)
        TweakInterface(('_literal', None), ('color', '#red'))

        And the value of the attribute just set is available like this:

        >>> abjad.tweak(markup).color
        '#red'

        Trying to get an attribute that has not yet been set raises an
        attribute error:

        >>> abjad.tweak(markup).Foo
        Traceback (most recent call last):
            ...
        AttributeError: TweakInterface object has no attribute 'Foo'.

    """

    ### INITIALIZER ###

    def __init__(
        self, *, deactivate: bool = None, literal: bool = None, tag: _tag.Tag = None
    ) -> None:
        if deactivate is not None:
            self._currently_deactivated = deactivate
        if literal is not None:
            literal = bool(literal)
        self._literal = literal
        if tag is not None:
            self._currently_tagging = tag

    ### SPECIAL METHODS ###

    def __getattr__(self, name) -> typing.Union[Interface, typing.Any]:
        r"""
        Gets Interface (or OverrideInterface) keyed to ``name``.

        ..  container:: example

            Tweaks may be tagged:

            >>> staff = abjad.Staff("c'4 d' e' f'")
            >>> markup = abjad.Markup(r"\italic Allegro", direction=abjad.Up)
            >>> abjad.tweak(markup, tag=abjad.Tag("+PARTS")).color = "#red"
            >>> abjad.attach(markup, staff[0])
            >>> abjad.show(staff) # doctest: +SKIP

            >>> string = abjad.lilypond(staff, tags=True)
            >>> print(string)
            \new Staff
            {
                c'4
                %! +PARTS
                - \tweak color #red
                ^ \markup {
                    \italic
                        Allegro
                    }
                d'4
                e'4
                f'4
            }

            Tweaks may be tagged with ``deactivate=True``:

            >>> staff = abjad.Staff("c'4 d' e' f'")
            >>> markup = abjad.Markup(r"\italic Allegro", direction=abjad.Up)
            >>> abjad.tweak(
            ...     markup, deactivate=True, tag=abjad.Tag("+PARTS")
            ... ).color = "#red"
            >>> abjad.attach(markup, staff[0])
            >>> abjad.show(staff) # doctest: +SKIP

            >>> string = abjad.lilypond(staff, tags=True)
            >>> print(string)
            \new Staff
            {
                c'4
                %! +PARTS
                - \tweak color #red
                ^ \markup {
                    \italic
                        Allegro
                    }
                d'4
                e'4
                f'4
            }

            Tweak tags and indicator tags may be set together:

            >>> staff = abjad.Staff("c'4 d' e' f'")
            >>> markup = abjad.Markup(r"\italic Allegro", direction=abjad.Up)
            >>> abjad.tweak(markup, tag=abjad.Tag("+PARTS")).color = "#red"
            >>> abjad.attach(markup, staff[0], tag=abjad.Tag("RED:M1"))
            >>> abjad.show(staff) # doctest: +SKIP

            >>> string = abjad.lilypond(staff, tags=True)
            >>> print(string)
            \new Staff
            {
                c'4
                %! +PARTS
                %! RED
                %! M1
                - \tweak color #red
                %! RED
                %! M1
                ^ \markup {
                %! RED
                %! M1
                    \italic
                %! RED
                %! M1
                        Allegro
                %! RED
                %! M1
                    }
                d'4
                e'4
                f'4
            }

        ..  container:: example

            Preloaded tweak managers can be made like this:

            >>> tweaks = abjad.TweakInterface()
            >>> tweaks.color = "#red"
            >>> tweaks.Y_offset = 6
            >>> tweaks
            TweakInterface(('Y_offset', 6), ('_literal', None), ('color', '#red'))

            Use the ``abjad.tweak()`` factory function for a shortcut:

            >>> tweaks = abjad.tweak("#red").color
            >>> tweaks
            TweakInterface(('_literal', None), ('color', '#red'))

            >>> tweaks.Y_offset = 6
            >>> tweaks
            TweakInterface(('Y_offset', 6), ('_literal', None), ('color', '#red'))

        ..  container:: example

            Set long LilyPond grob chains like this:

            >>> abjad.tweak(False).bound_details__left_broken__text
            TweakInterface(('_literal', None), ('bound_details__left_broken__text', False))

        """
        if name == "_currently_deactivated":
            return vars(self).get("_currently_deactivated")
        if name == "_currently_tagging":
            return vars(self).get("_currently_tagging")
        if name == "_literal":
            return vars(self).get("_literal")
        if "_pending_value" in vars(self):
            _pending_value = self._pending_value
            self.__setattr__(name, _pending_value)
            delattr(self, "_pending_value")
            return self
        camel_name = String(name).to_upper_camel_case()
        # assert name == camel_name, repr((name, camel_name))
        if name.startswith("_"):
            try:
                return vars(self)[name]
            except KeyError:
                type_name = type(self).__name__
                raise AttributeError(f"{type_name} object has no attribute {name!r}.")
        elif camel_name in grob_interfaces:
            try:
                return vars(self)[name]
            except KeyError:
                vars(self)[name] = Interface()
                return vars(self)[name]
        else:
            try:
                return vars(self)[name]
            except KeyError:
                type_name = type(self).__name__
                raise AttributeError(f"{type_name} object has no attribute {name!r}.")

    def __setattr__(self, name, value):
        """
        Sets attribute ``name`` equal to ``value``.

        ..  container:: example

            Allows LilyPond colors:

            >>> abjad.tweak("#blue").color
            TweakInterface(('_literal', None), ('color', '#blue'))

            >>> abjad.tweak("#(x11-color 'ForestGreen)").color
            TweakInterface(('_literal', None), ('color', "#(x11-color 'ForestGreen)"))

        """
        tag = getattr(self, "_currently_tagging", None)
        deactivate = getattr(self, "_currently_deactivated", None)
        if tag is not None:
            if deactivate is True:
                value = ("TAGGED", value, tag, True)
            else:
                value = ("TAGGED", value, tag)
        object.__setattr__(self, name, value)
        if name in ("_currently_deactivated", "_currently_tagging"):
            return
        try:
            delattr(self, "_currently_deactivated")
        except AttributeError:
            pass
        try:
            delattr(self, "_currently_tagging")
        except AttributeError:
            pass

    ### PRIVATE METHODS ###

    def _get_attribute_tuples(self) -> typing.List[typing.Tuple]:
        result: typing.List[typing.Tuple] = []
        for name, value in vars(self).items():
            if name == "_currently_tagging":
                continue
            if name == "_literal":
                continue
            if type(value) is Interface:
                grob_name = name
                grob_proxy = value
                pairs = iter(vars(grob_proxy).items())
                for attribute_name, attribute_value in pairs:
                    triple = (grob_name, attribute_name, attribute_value)
                    result.append(triple)
            else:
                attribute_name = name
                attribute_value = value
                pair = (attribute_name, attribute_value)
                result.append(pair)
        result.sort()
        return result

    def _list_format_contributions(self, directed=True):
        result = []
        for attribute_tuple in self._get_attribute_tuples():
            if len(attribute_tuple) == 2:
                grob = None
                attribute = attribute_tuple[0]
                value = attribute_tuple[1]
            elif len(attribute_tuple) == 3:
                grob = attribute_tuple[0]
                attribute = attribute_tuple[1]
                value = attribute_tuple[2]
            else:
                raise ValueError(f"invalid attribute tuple: {attribute_tuple!r}.")
            deactivate = False
            if isinstance(value, tuple) and value[0] == "TAGGED":
                if len(value) == 4:
                    deactivate = value[3]
                tag = value[2]
                value = value[1]
            else:
                tag = None
            string = make_lilypond_tweak_string(
                attribute,
                value,
                directed=directed,
                grob=grob,
                literal=self._literal,
            )
            strings = [string]
            if tag is not None:
                strings = _tag.double_tag(strings, tag, deactivate=deactivate)
            result.extend(strings)
        return result

    ### PUBLIC METHODS ###

    # TODO: move this somewhere clearer
    @staticmethod
    def set_tweaks(
        argument, manager: typing.Optional["TweakInterface"]
    ) -> typing.Optional["TweakInterface"]:
        r"""
        Sets tweaks on ``argument``.

        ..  container:: example

            >>> glissando = abjad.Glissando()
            >>> glissando.tweaks is None
            True

            >>> tweaks = abjad.tweak("blue").color
            >>> abjad.TweakInterface.set_tweaks(glissando, tweaks)
            TweakInterface(('_literal', None), ('color', 'blue'))

            >>> abjad.tweak(glissando)
            TweakInterface(('_literal', None), ('color', 'blue'))

        """
        if not hasattr(argument, "_tweaks"):
            try:
                argument._tweaks = None
            except AttributeError:
                name = type(argument).__name__
                raise NotImplementedError(f"{name} does not implement tweaks.")
        if manager is None:
            return None
        if not isinstance(manager, TweakInterface):
            raise Exception(f"must be tweak manager (not {manager!r}).")
        if argument._tweaks is None:
            argument._tweaks = TweakInterface(literal=manager._literal)
        existing_manager = argument._tweaks
        for tuple_ in manager._get_attribute_tuples():
            if len(tuple_) == 2:
                attribute, value = tuple_
                value = copy.copy(value)
                setattr(existing_manager, attribute, value)
            elif len(tuple_) == 3:
                grob, attribute, value = tuple_
                value = copy.copy(value)
                grob = getattr(existing_manager, grob)
                setattr(grob, attribute, value)
            else:
                message = "tweak tuple must have length 2 or 3"
                message += f" (not {tuple_!r})."
                raise ValueError(message)
        return existing_manager


IndexedTweakManager = typing.Union[TweakInterface, typing.Tuple[TweakInterface, int]]

IndexedTweakManagers = typing.Tuple[IndexedTweakManager, ...]


### FUNCTIONS ###


def tweak(argument, *, deactivate=None, expression=None, literal=None, tag=None):
    r"""
    Makes LilyPond tweak manager.

    ..  container:: example

        Tweaks markup:

        >>> staff = abjad.Staff("c'4 d' e' f'")
        >>> markup = abjad.Markup("Allegro assai", direction=abjad.Up)
        >>> abjad.tweak(markup).color = "#red"
        >>> abjad.attach(markup, staff[0])
        >>> abjad.show(staff) # doctest: +SKIP

        ..  docs::

            >>> string = abjad.lilypond(staff)
            >>> print(string)
            \new Staff
            {
                c'4
                - \tweak color #red
                ^ \markup { Allegro assai }
                d'4
                e'4
                f'4
            }

        Survives copy:

        >>> import copy
        >>> staff = abjad.Staff("c'4 d' e' f'")
        >>> markup_1 = abjad.Markup("Allegro assai", direction=abjad.Up)
        >>> abjad.tweak(markup_1).color = "#red"
        >>> markup_2 = copy.copy(markup_1)
        >>> abjad.attach(markup_2, staff[0])
        >>> abjad.show(staff) # doctest: +SKIP

        ..  docs::

            >>> string = abjad.lilypond(staff)
            >>> print(string)
            \new Staff
            {
                c'4
                - \tweak color #red
                ^ \markup { Allegro assai }
                d'4
                e'4
                f'4
            }

        Survives dot-chaining:

        >>> staff = abjad.Staff("c'4 d' e' f'")
        >>> markup = abjad.Markup(r'\italic "Allegro assai"', direction=abjad.Up)
        >>> abjad.tweak(markup).color = "#red"
        >>> abjad.attach(markup, staff[0])
        >>> abjad.show(staff) # doctest: +SKIP

        ..  docs::

            >>> string = abjad.lilypond(staff)
            >>> print(string)
            \new Staff
            {
                c'4
                - \tweak color #red
                ^ \markup {
                    \italic
                        "Allegro assai"
                    }
                d'4
                e'4
                f'4
            }

        Works for opposite-directed coincident markup:

        >>> staff = abjad.Staff("c'4 d' e' f'")
        >>> markup_1 = abjad.Markup("Allegro assai ...", direction=abjad.Up)
        >>> abjad.tweak(markup_1).color = "#red"
        >>> abjad.attach(markup_1, staff[0])
        >>> markup_2 = abjad.Markup("... ma non troppo", direction=abjad.Down)
        >>> abjad.tweak(markup_2).color = "#blue"
        >>> abjad.tweak(markup_2).staff_padding = 4
        >>> abjad.attach(markup_2, staff[0])
        >>> abjad.show(staff) # doctest: +SKIP

        ..  docs::

            >>> string = abjad.lilypond(staff)
            >>> print(string)
            \new Staff
            {
                c'4
                - \tweak color #red
                ^ \markup { Allegro assai ... }
                - \tweak color #blue
                - \tweak staff-padding 4
                _ \markup { ... ma non troppo }
                d'4
                e'4
                f'4
            }

        Ignored for same-directed coincident markup:

        >>> staff = abjad.Staff("c'4 d' e' f'")
        >>> markup_1 = abjad.Markup("Allegro assai ...", direction=abjad.Up)
        >>> abjad.tweak(markup_1).color = "#red"
        >>> abjad.attach(markup_1, staff[0])
        >>> markup_2 = abjad.Markup("... ma non troppo", direction=abjad.Up)
        >>> abjad.tweak(markup_2).color = "#blue"
        >>> abjad.tweak(markup_2).staff_padding = 4
        >>> abjad.attach(markup_2, staff[0])
        >>> abjad.show(staff) # doctest: +SKIP

        ..  docs::

            >>> string = abjad.lilypond(staff)
            >>> print(string)
            \new Staff
            {
                c'4
                - \tweak color #red
                ^ \markup { Allegro assai ... }
                - \tweak color #blue
                - \tweak staff-padding 4
                ^ \markup { ... ma non troppo }
                d'4
                e'4
                f'4
            }

    ..  container:: example

        Tweaks note-head:

        >>> staff = abjad.Staff("c'4 cs' d' ds'")
        >>> abjad.tweak(staff[1].note_head).color = "#red"
        >>> abjad.show(staff) # doctest: +SKIP

        ..  docs::

            >>> string = abjad.lilypond(staff)
            >>> print(string)
            \new Staff
            {
                c'4
                \tweak color #red
                cs'4
                d'4
                ds'4
            }

        Tweaks grob aggregated to note-head:

        >>> staff = abjad.Staff("c'4 cs' d' ds'")
        >>> abjad.tweak(staff[1].note_head).Accidental.color = "#red"
        >>> abjad.show(staff) # doctest: +SKIP

        ..  docs::

            >>> string = abjad.lilypond(staff)
            >>> print(string)
            \new Staff
            {
                c'4
                \tweak Accidental.color #red
                cs'4
                d'4
                ds'4
            }

    ..  container:: example

        Tweaks can be tagged:

        >>> staff = abjad.Staff("c'4 d' e' f'")
        >>> dynamic = abjad.Dynamic("f")
        >>> abjad.tweak(dynamic, tag=abjad.Tag("RED")).color = "#red"
        >>> abjad.attach(dynamic, staff[0])

        >>> string = abjad.lilypond(staff, tags=True)
        >>> print(string)
        \new Staff
        {
            c'4
            %! RED
            - \tweak color #red
            \f
            d'4
            e'4
            f'4
        }

        >>> abjad.show(staff) # doctest: +SKIP

        REGRESSION. Tweaked tags can be set multiple times:

        >>> staff = abjad.Staff("c'4 d' e' f'")
        >>> dynamic = abjad.Dynamic("f")
        >>> abjad.tweak(dynamic, tag=abjad.Tag("RED")).color = "#red"
        >>> abjad.tweak(dynamic, tag=abjad.Tag("BLUE")).color = "#blue"
        >>> abjad.attach(dynamic, staff[0])

        >>> string = abjad.lilypond(staff, tags=True)
        >>> print(string)
        \new Staff
        {
            c'4
            %! BLUE
            - \tweak color #blue
            \f
            d'4
            e'4
            f'4
        }

        >>> abjad.show(staff) # doctest: +SKIP

    ..  container:: example

        Returns LilyPond tweak manager:

        >>> abjad.tweak(markup_1)
        TweakInterface(('_literal', None), ('color', '#red'))

    ..  container:: example

        Tweak expressions work like this:

        >>> abjad.tweak("#red").color
        TweakInterface(('_literal', None), ('color', '#red'))

        >>> abjad.tweak(6).Y_offset
        TweakInterface(('Y_offset', 6), ('_literal', None))

        >>> abjad.tweak(False).bound_details__left_broken__text
        TweakInterface(('_literal', None), ('bound_details__left_broken__text', False))

    """
    if tag is not None and not isinstance(tag, _tag.Tag):
        raise Exception(f"must be be tag: {repr(tag)}")

    constants = (enums.Down, enums.Left, enums.Right, enums.Up)
    prototype = (bool, int, float, str, tuple)
    if expression is True or argument in constants or isinstance(argument, prototype):
        interface = TweakInterface(deactivate=deactivate, literal=literal, tag=tag)
        interface._pending_value = argument
        return interface
    if not hasattr(argument, "_tweaks"):
        name = type(argument).__name__
        raise NotImplementedError(f"{name} does not allow tweaks (yet).")
    if argument._tweaks is None:
        interface = TweakInterface(deactivate=deactivate, literal=literal, tag=tag)
        argument._tweaks = interface
    else:
        interface = argument._tweaks
        interface.__init__(deactivate=deactivate, literal=literal, tag=tag)
    return interface



#! /usr/bin/env python
import os
import sys

import setuptools

CURRENT_PYTHON = sys.version_info[:2]
REQUIRED_PYTHON = (3, 6)

if CURRENT_PYTHON < REQUIRED_PYTHON:
    sys.stderr.write(
        """
==========================
Unsupported Python version
==========================

This version of Abjad requires Python {}.{}, but you're trying to
install it on Python {}.{}.

This may be because you are using a version of pip that doesn't
understand the python_requires classifier. Make sure you
have pip >= 9.0 and setuptools >= 24.2, then try again:

    $ python -m pip install --upgrade pip setuptools
    $ python -m pip install abjad

This will install the latest version of Abjad which works on your
version of Python.
""".format(
            *(REQUIRED_PYTHON + CURRENT_PYTHON)
        )
    )
    sys.exit(1)

version_file_path = os.path.join(os.path.dirname(__file__), "abjad", "_version.py")
with open(version_file_path, "r") as file_pointer:
    file_contents_string = file_pointer.read()
local_dict: dict = {}
exec(file_contents_string, None, local_dict)
__version__ = local_dict["__version__"]

description = "Abjad is a Python API for building LilyPond files."

with open("README.rst", "r") as file_pointer:
    long_description = file_pointer.read()

author = ["Trevor Bača", "Josiah Wolf Oberholtzer"]

author_email = [
    "trevor.baca@gmail.com",
    "josiah.oberholtzer@gmail.com",
]

classifiers = [
    "Development Status :: 5 - Production/Stable",
    "License :: OSI Approved :: GNU General Public License (GPL)",
    "Programming Language :: Python :: 3.6",
    "Programming Language :: Python :: 3.7",
    "Programming Language :: Python :: 3.8",
    "Programming Language :: Python :: Implementation :: CPython",
    "Topic :: Artistic Software",
]

extras_require = {
    "nauert": ["abjad-ext-nauert==3.4"],
    "rmakers": ["abjad-ext-rmakers==3.4"],
}

keywords = [
    "music composition",
    "music notation",
    "lilypond",
]

install_requires = [
    "black>=20.8b1",
    "flake8",
    "isort",
    "mypy>=0.770",
    "ply",
    "pytest>=5.4.3",
    "pytest-cov>=2.7.1",
    "pytest-helpers-namespace",
    "quicktions>=1.3",
    "roman",
    "six",
    "sphinx-autodoc-typehints",
    "uqbar>=0.4.4, <0.5.0",
]

if __name__ == "__main__":
    setuptools.setup(
        author=", ".join(author),
        author_email=", ".join(author_email),
        classifiers=classifiers,
        description=description,
        extras_require=extras_require,
        include_package_data=True,
        install_requires=install_requires,
        keywords=", ".join(keywords),
        license="MIT",
        long_description=long_description,
        long_description_content_type="text/x-rst",
        name="abjad",
        packages=["abjad"],
        platforms="Any",
        url="https://abjad.github.io",
        version=__version__,
    )



import typing

from .. import enums
from ..bundle import LilyPondFormatBundle
from ..overrides import TweakInterface
from ..storage import StorageFormatManager


class LaissezVibrer:
    r"""
    Laissez vibrer.

    ..  container:: example

        >>> chord = abjad.Chord("<c' e' g' c''>4")
        >>> laissez_vibrer = abjad.LaissezVibrer()
        >>> abjad.attach(laissez_vibrer, chord)
        >>> abjad.show(chord) # doctest: +SKIP

        ..  docs::

            >>> string = abjad.lilypond(chord)
            >>> print(string)
            <c' e' g' c''>4
            \laissezVibrer

    """

    ### CLASS VARIABLES ###

    __slots__ = ("_tweaks",)

    _format_slot = "after"

    _time_orientation = enums.Right

    ### INITIALIZER ###

    def __init__(self, *, tweaks: TweakInterface = None) -> None:
        if tweaks is not None:
            assert isinstance(tweaks, TweakInterface), repr(tweaks)
        self._tweaks = TweakInterface.set_tweaks(self, tweaks)

    ### SPECIAL METHODS ###

    def __eq__(self, argument) -> bool:
        """
        Is true when all initialization values of Abjad value object equal
        the initialization values of ``argument``.
        """
        return StorageFormatManager.compare_objects(self, argument)

    def __hash__(self) -> int:
        """
        Hashes Abjad value object.
        """
        hash_values = StorageFormatManager(self).get_hash_values()
        try:
            result = hash(hash_values)
        except TypeError:
            raise TypeError(f"unhashable type: {self}")
        return result

    def __repr__(self) -> str:
        """
        Gets interpreter representation.
        """
        return StorageFormatManager(self).get_repr_format()

    def __str__(self) -> str:
        r"""
        Gets string representation of laissez vibrer indicator.

        ..  container:: example

            Default:

            >>> str(abjad.LaissezVibrer())
            '\\laissezVibrer'

        """
        return r"\laissezVibrer"

    ### PRIVATE METHODS ###

    def _get_lilypond_format(self):
        return str(self)

    def _get_lilypond_format_bundle(self, component=None):
        bundle = LilyPondFormatBundle()
        if self.tweaks:
            tweaks = self.tweaks._list_format_contributions()
            bundle.after.articulations.extend(tweaks)
        bundle.after.articulations.append(self._get_lilypond_format())
        return bundle

    ### PUBLIC PROPERTIES ###

    @property
    def tweaks(self) -> typing.Optional[TweakInterface]:
        r"""
        Gets tweaks

        ..  container:: example

            >>> note = abjad.Note("c'4")
            >>> lv = abjad.LaissezVibrer()
            >>> abjad.tweak(lv).color = "#blue"
            >>> abjad.attach(lv, note)
            >>> abjad.show(note) # doctest: +SKIP

            ..  docs::

                >>> string = abjad.lilypond(note)
                >>> print(string)
                c'4
                - \tweak color #blue
                \laissezVibrer

        """
        return self._tweaks



from quicktions import Fraction

from . import (
    deprecated,
    enumerate,
    get,
    illustrators,
    io,
    iterpitches,
    lyconst,
    lyenv,
    makers,
    mutate,
    persist,
    wf,
)
from ._version import __version__, __version_info__
from .attach import Wrapper, annotate, attach, detach
from .bundle import LilyPondFormatBundle, SlotContributions
from .configuration import (
    Configuration,
    list_all_classes,
    list_all_functions,
    yield_all_modules,
)
from .contextmanagers import (
    ContextManager,
    FilesystemState,
    ForbidUpdate,
    NullContextManager,
    ProgressIndicator,
    RedirectedStreams,
    TemporaryDirectory,
    TemporaryDirectoryChange,
    Timer,
)
from .cyclictuple import CyclicTuple
from .duration import Duration, Multiplier, NonreducedFraction, Offset
from .enums import (
    Center,
    Comparison,
    Down,
    Exact,
    HorizontalAlignment,
    Left,
    Less,
    Middle,
    More,
    Right,
    Up,
    VerticalAlignment,
)
from .exceptions import (
    AssignabilityError,
    ImpreciseMetronomeMarkError,
    LilyPondParserError,
    MissingMetronomeMarkError,
    ParentageError,
    PersistentIndicatorError,
    SchemeParserFinishedError,
    UnboundedTimeIntervalError,
    WellformednessError,
)
from .expression import Expression
from .format import LilyPondFormatManager
from .get import Descendants, Lineage
from .illustrators import illustrate
from .indicators.Arpeggio import Arpeggio
from .indicators.Articulation import Articulation
from .indicators.BarLine import BarLine
from .indicators.BeamCount import BeamCount
from .indicators.BendAfter import BendAfter
from .indicators.BowContactPoint import BowContactPoint
from .indicators.BowMotionTechnique import BowMotionTechnique
from .indicators.BowPressure import BowPressure
from .indicators.BreathMark import BreathMark
from .indicators.Clef import Clef, StaffPosition
from .indicators.ColorFingering import ColorFingering
from .indicators.Dynamic import Dynamic
from .indicators.Fermata import Fermata
from .indicators.Glissando import Glissando
from .indicators.KeyCluster import KeyCluster
from .indicators.KeySignature import KeySignature
from .indicators.LaissezVibrer import LaissezVibrer
from .indicators.LilyPondComment import LilyPondComment
from .indicators.MarginMarkup import MarginMarkup
from .indicators.MetronomeMark import MetronomeMark
from .indicators.Mode import Mode
from .indicators.Ottava import Ottava
from .indicators.RehearsalMark import RehearsalMark
from .indicators.Repeat import Repeat
from .indicators.RepeatTie import RepeatTie
from .indicators.StaffChange import StaffChange
from .indicators.StartBeam import StartBeam
from .indicators.StartGroup import StartGroup
from .indicators.StartHairpin import StartHairpin
from .indicators.StartMarkup import StartMarkup
from .indicators.StartPhrasingSlur import StartPhrasingSlur
from .indicators.StartPianoPedal import StartPianoPedal
from .indicators.StartSlur import StartSlur
from .indicators.StartTextSpan import StartTextSpan
from .indicators.StartTrillSpan import StartTrillSpan
from .indicators.StemTremolo import StemTremolo
from .indicators.StopBeam import StopBeam
from .indicators.StopGroup import StopGroup
from .indicators.StopHairpin import StopHairpin
from .indicators.StopPhrasingSlur import StopPhrasingSlur
from .indicators.StopPianoPedal import StopPianoPedal
from .indicators.StopSlur import StopSlur
from .indicators.StopTextSpan import StopTextSpan
from .indicators.StopTrillSpan import StopTrillSpan
from .indicators.StringContactPoint import StringContactPoint
from .indicators.Tie import Tie
from .indicators.TimeSignature import TimeSignature
from .instruments import (
    Accordion,
    AltoFlute,
    AltoSaxophone,
    AltoTrombone,
    AltoVoice,
    BaritoneSaxophone,
    BaritoneVoice,
    BassClarinet,
    BassFlute,
    BassSaxophone,
    BassTrombone,
    BassVoice,
    Bassoon,
    Cello,
    ClarinetInA,
    ClarinetInBFlat,
    ClarinetInEFlat,
    Contrabass,
    ContrabassClarinet,
    ContrabassFlute,
    ContrabassSaxophone,
    Contrabassoon,
    EnglishHorn,
    Flute,
    FrenchHorn,
    Glockenspiel,
    Guitar,
    Harp,
    Harpsichord,
    Instrument,
    Marimba,
    MezzoSopranoVoice,
    Oboe,
    Percussion,
    Piano,
    Piccolo,
    SopraninoSaxophone,
    SopranoSaxophone,
    SopranoVoice,
    StringNumber,
    TenorSaxophone,
    TenorTrombone,
    TenorVoice,
    Trumpet,
    Tuba,
    Tuning,
    Vibraphone,
    Viola,
    Violin,
    Xylophone,
)
from .io import graph, play, show
from .iterate import Iteration, iterate
from .label import ColorMap, Label
from .lilypond import lilypond
from .lilypondfile import (
    Block,
    ContextBlock,
    DateTimeToken,
    LilyPondDimension,
    LilyPondFile,
    LilyPondLanguageToken,
    LilyPondVersionToken,
    PackageGitCommitToken,
)
from .lyproxy import (
    LilyPondContext,
    LilyPondEngraver,
    LilyPondGrob,
    LilyPondGrobInterface,
)
from .makers import LeafMaker, NoteMaker
from .markups import Markup, Postscript, PostscriptOperator
from .math import Infinity, NegativeInfinity
from .meter import Meter, MeterList, MetricAccentKernel, OffsetCounter
from .metricmodulation import MetricModulation
from .new import new
from .obgc import OnBeatGraceContainer, on_beat_grace_container
from .ordereddict import OrderedDict
from .overrides import (
    IndexedTweakManager,
    IndexedTweakManagers,
    Interface,
    LilyPondLiteral,
    LilyPondOverride,
    LilyPondSetting,
    OverrideInterface,
    SettingInterface,
    TweakInterface,
    override,
    setting,
    tweak,
)
from .parentage import Parentage
from .parsers import parser
from .parsers.base import Parser
from .parsers.parse import parse
from .pattern import Pattern, PatternTuple
from .pitch.Accidental import Accidental
from .pitch.Octave import Octave
from .pitch.PitchRange import PitchRange
from .pitch.SetClass import SetClass
from .pitch.intervalclasses import (
    IntervalClass,
    NamedIntervalClass,
    NamedInversionEquivalentIntervalClass,
    NumberedIntervalClass,
    NumberedInversionEquivalentIntervalClass,
)
from .pitch.intervals import Interval, NamedInterval, NumberedInterval
from .pitch.operators import (
    CompoundOperator,
    Duplication,
    Inversion,
    Multiplication,
    Retrograde,
    Rotation,
    Transposition,
)
from .pitch.pitchclasses import NamedPitchClass, NumberedPitchClass, PitchClass
from .pitch.pitches import NamedPitch, NumberedPitch, Pitch, PitchTyping
from .pitch.segments import (
    IntervalClassSegment,
    IntervalSegment,
    PitchClassSegment,
    PitchSegment,
    Segment,
    TwelveToneRow,
)
from .pitch.sets import IntervalClassSet, IntervalSet, PitchClassSet, PitchSet, Set
from .pitch.vectors import (
    IntervalClassVector,
    IntervalVector,
    PitchClassVector,
    PitchVector,
    Vector,
)
from .ratio import NonreducedRatio, Ratio
from .score import (
    AfterGraceContainer,
    BeforeGraceContainer,
    Chord,
    Cluster,
    Component,
    Container,
    Context,
    DrumNoteHead,
    Leaf,
    MultimeasureRest,
    Note,
    NoteHead,
    NoteHeadList,
    Rest,
    Score,
    Skip,
    Staff,
    StaffGroup,
    TremoloContainer,
    Tuplet,
    Voice,
)
from .segmentmaker import SegmentMaker
from .select import (
    DurationInequality,
    Inequality,
    LengthInequality,
    LogicalTie,
    PitchInequality,
    Selection,
    select,
)
from .sequence import Sequence
from .spanners import (
    beam,
    bow_contact_spanner,
    glissando,
    hairpin,
    horizontal_bracket,
    ottava,
    phrasing_slur,
    piano_pedal,
    slur,
    text_spanner,
    tie,
    trill_spanner,
)
from .storage import (
    FormatSpecification,
    StorageFormatManager,
    StorageFormatSpecification,
    storage,
)
from .string import String
from .tag import Line, Tag, activate, deactivate
from .templates import (
    GroupedRhythmicStavesScoreTemplate,
    GroupedStavesScoreTemplate,
    ScoreTemplate,
    StringOrchestraScoreTemplate,
    StringQuartetScoreTemplate,
    TwoStaffPianoScoreTemplate,
)
from .timespan import AnnotatedTimespan, Timespan, TimespanList
from .typedcollections import (
    TypedCollection,
    TypedCounter,
    TypedFrozenset,
    TypedList,
    TypedTuple,
)
from .typings import (
    DurationSequenceTyping,
    DurationTyping,
    IntegerPair,
    IntegerSequence,
    Number,
    NumberPair,
    PatternTyping,
    Prototype,
    RatioSequenceTyping,
    RatioTyping,
    Strings,
)
from .verticalmoment import (
    VerticalMoment,
    iterate_leaf_pairs,
    iterate_pitch_pairs,
    iterate_vertical_moments,
)

index = Pattern.index
index_all = Pattern.index_all
index_first = Pattern.index_first
index_last = Pattern.index_last


__all__ = [
    "Accidental",
    "Accordion",
    "AfterGraceContainer",
    "AltoFlute",
    "AltoSaxophone",
    "AltoTrombone",
    "AltoVoice",
    "AnnotatedTimespan",
    "Arpeggio",
    "Articulation",
    "AssignabilityError",
    "BarLine",
    "BaritoneSaxophone",
    "BaritoneVoice",
    "BassClarinet",
    "BassFlute",
    "BassSaxophone",
    "BassTrombone",
    "BassVoice",
    "Bassoon",
    "BeamCount",
    "BeforeGraceContainer",
    "BendAfter",
    "Block",
    "BowContactPoint",
    "BowMotionTechnique",
    "BowPressure",
    "BreathMark",
    "Cello",
    "Center",
    "Chord",
    "ClarinetInA",
    "ClarinetInBFlat",
    "ClarinetInEFlat",
    "Clef",
    "Cluster",
    "ColorFingering",
    "ColorMap",
    "Comparison",
    "Component",
    "CompoundOperator",
    "Configuration",
    "Container",
    "Context",
    "ContextBlock",
    "ContextManager",
    "Contrabass",
    "ContrabassClarinet",
    "ContrabassFlute",
    "ContrabassSaxophone",
    "Contrabassoon",
    "CyclicTuple",
    "DateTimeToken",
    "Descendants",
    "Down",
    "DrumNoteHead",
    "Duplication",
    "Duration",
    "DurationInequality",
    "DurationSequenceTyping",
    "DurationTyping",
    "Dynamic",
    "EnglishHorn",
    "Exact",
    "Expression",
    "Fermata",
    "FilesystemState",
    "Flute",
    "ForbidUpdate",
    "FormatSpecification",
    "Fraction",
    "FrenchHorn",
    "Glissando",
    "Glockenspiel",
    "GroupedRhythmicStavesScoreTemplate",
    "GroupedStavesScoreTemplate",
    "Guitar",
    "Harp",
    "Harpsichord",
    "HorizontalAlignment",
    "ImpreciseMetronomeMarkError",
    "IndexedTweakManager",
    "IndexedTweakManagers",
    "Inequality",
    "Infinity",
    "Instrument",
    "IntegerPair",
    "IntegerSequence",
    "Interface",
    "Interval",
    "IntervalClass",
    "IntervalClassSegment",
    "IntervalClassSet",
    "IntervalClassVector",
    "IntervalSegment",
    "IntervalSet",
    "IntervalVector",
    "Inversion",
    "Iteration",
    "KeyCluster",
    "KeySignature",
    "Label",
    "LaissezVibrer",
    "Leaf",
    "LeafMaker",
    "Left",
    "LengthInequality",
    "Less",
    "LilyPondComment",
    "LilyPondContext",
    "LilyPondDimension",
    "LilyPondEngraver",
    "LilyPondFile",
    "LilyPondFormatBundle",
    "LilyPondFormatManager",
    "LilyPondGrob",
    "LilyPondGrobInterface",
    "LilyPondLanguageToken",
    "LilyPondLiteral",
    "LilyPondOverride",
    "LilyPondParserError",
    "LilyPondSetting",
    "LilyPondVersionToken",
    "Line",
    "Lineage",
    "LogicalTie",
    "MarginMarkup",
    "Marimba",
    "Markup",
    "Meter",
    "MeterList",
    "MetricAccentKernel",
    "MetricModulation",
    "MetronomeMark",
    "MezzoSopranoVoice",
    "Middle",
    "MissingMetronomeMarkError",
    "Mode",
    "More",
    "MultimeasureRest",
    "Multiplication",
    "Multiplier",
    "NamedInterval",
    "NamedIntervalClass",
    "NamedInversionEquivalentIntervalClass",
    "NamedPitch",
    "NamedPitchClass",
    "NegativeInfinity",
    "NonreducedFraction",
    "NonreducedRatio",
    "Note",
    "NoteHead",
    "NoteHeadList",
    "NoteMaker",
    "NullContextManager",
    "Number",
    "NumberPair",
    "NumberedInterval",
    "NumberedIntervalClass",
    "NumberedInversionEquivalentIntervalClass",
    "NumberedPitch",
    "NumberedPitchClass",
    "Oboe",
    "Octave",
    "Offset",
    "OffsetCounter",
    "OnBeatGraceContainer",
    "OrderedDict",
    "Ottava",
    "OverrideInterface",
    "PackageGitCommitToken",
    "Parentage",
    "ParentageError",
    "Parser",
    "Pattern",
    "PatternTuple",
    "PatternTyping",
    "Percussion",
    "PersistentIndicatorError",
    "Piano",
    "Piccolo",
    "Pitch",
    "PitchClass",
    "PitchClassSegment",
    "PitchClassSet",
    "PitchClassVector",
    "PitchInequality",
    "PitchRange",
    "PitchSegment",
    "PitchSet",
    "PitchTyping",
    "PitchVector",
    "Postscript",
    "PostscriptOperator",
    "ProgressIndicator",
    "Prototype",
    "Ratio",
    "RatioSequenceTyping",
    "RatioTyping",
    "RedirectedStreams",
    "RehearsalMark",
    "Repeat",
    "RepeatTie",
    "Rest",
    "Retrograde",
    "Right",
    "Rotation",
    "SchemeParserFinishedError",
    "Score",
    "ScoreTemplate",
    "Segment",
    "SegmentMaker",
    "Selection",
    "Sequence",
    "Set",
    "SetClass",
    "SettingInterface",
    "Skip",
    "SlotContributions",
    "SopraninoSaxophone",
    "SopranoSaxophone",
    "SopranoVoice",
    "Staff",
    "StaffChange",
    "StaffGroup",
    "StaffPosition",
    "StartBeam",
    "StartGroup",
    "StartHairpin",
    "StartMarkup",
    "StartPhrasingSlur",
    "StartPianoPedal",
    "StartSlur",
    "StartTextSpan",
    "StartTrillSpan",
    "StemTremolo",
    "StopBeam",
    "StopGroup",
    "StopHairpin",
    "StopPhrasingSlur",
    "StopPianoPedal",
    "StopSlur",
    "StopTextSpan",
    "StopTrillSpan",
    "StorageFormatManager",
    "StorageFormatSpecification",
    "String",
    "StringContactPoint",
    "StringNumber",
    "StringOrchestraScoreTemplate",
    "StringQuartetScoreTemplate",
    "Strings",
    "Tag",
    "TemporaryDirectory",
    "TemporaryDirectoryChange",
    "TenorSaxophone",
    "TenorTrombone",
    "TenorVoice",
    "Tie",
    "TimeSignature",
    "Timer",
    "Timespan",
    "TimespanList",
    "Transposition",
    "TremoloContainer",
    "Trumpet",
    "Tuba",
    "Tuning",
    "Tuplet",
    "TweakInterface",
    "TwelveToneRow",
    "TwoStaffPianoScoreTemplate",
    "TypedCollection",
    "TypedCounter",
    "TypedFrozenset",
    "TypedList",
    "TypedTuple",
    "UnboundedTimeIntervalError",
    "Up",
    "Vector",
    "VerticalAlignment",
    "VerticalMoment",
    "Vibraphone",
    "Viola",
    "Violin",
    "Voice",
    "WellformednessError",
    "Wrapper",
    "Xylophone",
    "__version__",
    "__version_info__",
    "activate",
    "annotate",
    "attach",
    "beam",
    "bow_contact_spanner",
    "deactivate",
    "deprecated",
    "detach",
    "enumerate",
    "glissando",
    "graph",
    "hairpin",
    "horizontal_bracket",
    "illustrate",
    "illustrators",
    "index",
    "index_all",
    "index_first",
    "index_last",
    "get",
    "io",
    "iterate",
    "iterate_leaf_pairs",
    "iterate_pitch_pairs",
    "iterate_vertical_moments",
    "iterpitches",
    "lilypond",
    "list_all_classes",
    "list_all_functions",
    "lyconst",
    "lyenv",
    "makers",
    "mutate",
    "new",
    "on_beat_grace_container",
    "ottava",
    "override",
    "parse",
    "parser",
    "persist",
    "phrasing_slur",
    "piano_pedal",
    "pitch_class_segment",
    "play",
    "select",
    "setting",
    "show",
    "slur",
    "storage",
    "text_spanner",
    "tie",
    "trill_spanner",
    "tweak",
    "wf",
    "yield_all_modules",
]



import functools
import numbers
import typing

from ..bundle import LilyPondFormatBundle
from ..pitch import _lib as pitch__lib
from ..pitch.pitches import NamedPitch
from ..storage import FormatSpecification, StorageFormatManager


class Clef:
    r"""
    Clef.

    ..  container:: example

        Some available clefs:

        >>> staff = abjad.Staff("c'8 d'8 e'8 f'8 g'8 a'8 b'8 c''8")
        >>> clef = abjad.Clef('treble')
        >>> abjad.attach(clef, staff[0])
        >>> clef = abjad.Clef('alto')
        >>> abjad.attach(clef, staff[1])
        >>> clef = abjad.Clef('bass')
        >>> abjad.attach(clef, staff[2])
        >>> clef = abjad.Clef('treble^8')
        >>> abjad.attach(clef, staff[3])
        >>> clef = abjad.Clef('bass_8')
        >>> abjad.attach(clef, staff[4])
        >>> clef = abjad.Clef('tenor')
        >>> abjad.attach(clef, staff[5])
        >>> clef = abjad.Clef('bass^15')
        >>> abjad.attach(clef, staff[6])
        >>> clef = abjad.Clef('percussion')
        >>> abjad.attach(clef, staff[7])
        >>> abjad.show(staff) # doctest: +SKIP

        ..  docs::

            >>> string = abjad.lilypond(staff)
            >>> print(string)
            \new Staff
            {
                \clef "treble"
                c'8
                \clef "alto"
                d'8
                \clef "bass"
                e'8
                \clef "treble^8"
                f'8
                \clef "bass_8"
                g'8
                \clef "tenor"
                a'8
                \clef "bass^15"
                b'8
                \clef "percussion"
                c''8
            }

    ..  container:: example

        Clefs can be tagged:

        >>> staff = abjad.Staff("c'4 d' e' f'")
        >>> abjad.attach(
        ...     abjad.Clef('treble'), staff[0], tag=abjad.Tag("+PARTS")
        ... )
        >>> abjad.show(staff) # doctest: +SKIP

        >>> string = abjad.lilypond(staff, tags=True)
        >>> print(string)
        \new Staff
        {
            %! +PARTS
            \clef "treble"
            c'4
            d'4
            e'4
            f'4
        }

    ..  container:: example

        LilyPond can not handle simultaneous clefs:

        >>> voice_1 = abjad.Voice("e'8 g' f' a' g' b'")
        >>> abjad.attach(abjad.Clef('treble'), voice_1[0], context='Voice')
        >>> literal = abjad.LilyPondLiteral(r"\voiceOne", "opening")
        >>> abjad.attach(literal, voice_1)
        >>> voice_1.consists_commands.append('Clef_engraver')
        >>> voice_2 = abjad.Voice("c'4. c,8 b,, a,,")
        >>> abjad.attach(abjad.Clef('treble'), voice_2[0], context='Voice')
        >>> abjad.attach(abjad.Clef('bass'), voice_2[1], context='Voice')
        >>> literal = abjad.LilyPondLiteral(r"\voiceTwo", "opening")
        >>> abjad.attach(literal, voice_2)
        >>> voice_2.consists_commands.append('Clef_engraver')
        >>> staff = abjad.Staff([voice_1, voice_2], simultaneous=True)
        >>> staff.remove_commands.append('Clef_engraver')
        >>> abjad.show(staff) # doctest: +SKIP

        ..  docs::

            >>> string = abjad.lilypond(staff)
            >>> print(string)
            \new Staff
            \with
            {
                \remove Clef_engraver
            }
            <<
                \new Voice
                \with
                {
                    \consists Clef_engraver
                }
                {
                    \voiceOne
                    \clef "treble"
                    e'8
                    g'8
                    f'8
                    a'8
                    g'8
                    b'8
                }
                \new Voice
                \with
                {
                    \consists Clef_engraver
                }
                {
                    \voiceTwo
                    \clef "treble"
                    c'4.
                    \clef "bass"
                    c,8
                    b,,8
                    a,,8
                }
            >>

        But Abjad components work fine:

        >>> for leaf in abjad.select(voice_1).leaves():
        ...     leaf, abjad.get.effective(leaf, abjad.Clef)
        ...
        (Note("e'8"), Clef('treble'))
        (Note("g'8"), Clef('treble'))
        (Note("f'8"), Clef('treble'))
        (Note("a'8"), Clef('treble'))
        (Note("g'8"), Clef('treble'))
        (Note("b'8"), Clef('treble'))

        >>> for leaf in abjad.select(voice_2).leaves():
        ...     leaf, abjad.get.effective(leaf, abjad.Clef)
        ...
        (Note("c'4."), Clef('treble'))
        (Note('c,8'), Clef('bass'))
        (Note('b,,8'), Clef('bass'))
        (Note('a,,8'), Clef('bass'))

    """

    ### CLASS VARIABLES ###

    __slots__ = ("_hide", "_middle_c_position", "_name")

    _clef_name_to_middle_c_position = {
        "treble": -6,
        "alto": 0,
        "varC": 0,
        "tenor": 2,
        "tenorvarC": 2,
        "bass": 6,
        "french": -8,
        "soprano": -4,
        "mezzosoprano": -2,
        "baritone": 4,
        "varbaritone": 4,
        "percussion": 0,
        "tab": 0,
    }

    _context = "Staff"

    _format_slot = "opening"

    _persistent = True

    _redraw = True

    _to_width = {
        "alto": 2.75,
        "varC": 2.75,
        "bass": 2.75,
        "percussion": 2.5,
        "tenor": 2.75,
        "tenorvarC": 2.75,
        "treble": 2.5,
    }

    ### INITIALIZER ###

    def __init__(self, name: str = "treble", *, hide: bool = None) -> None:
        if isinstance(name, str):
            self._name = name
        elif isinstance(name, type(self)):
            self._name = name.name
        else:
            raise TypeError("can not initialize clef: {name!r}.")
        if hide is not None:
            hide = bool(hide)
        self._hide = hide
        middle_c_position = self._calculate_middle_c_position(self._name)
        middle_c_position = StaffPosition(middle_c_position)
        self._middle_c_position = middle_c_position

    ### SPECIAL METHODS ###

    def __eq__(self, argument) -> bool:
        """
        Is true when all initialization values of Abjad value object equal
        the initialization values of ``argument``.
        """
        return StorageFormatManager.compare_objects(self, argument)

    def __hash__(self) -> int:
        """
        Hashes Abjad value object.
        """
        hash_values = StorageFormatManager(self).get_hash_values()
        try:
            result = hash(hash_values)
        except TypeError:
            raise TypeError(f"unhashable type: {self}")
        return result

    def __repr__(self) -> str:
        """
        Gets interpreter representation.
        """
        return StorageFormatManager(self).get_repr_format()

    ### PRIVATE METHODS ###

    def _calculate_middle_c_position(self, clef_name):
        alteration = 0
        if "_" in self._name:
            base_name, part, suffix = clef_name.partition("_")
            if suffix == "8":
                alteration = 7
            elif suffix == "15":
                alteration = 13
            else:
                raise Exception(f"bad clef alteration suffix: {suffix!r}.")
        elif "^" in self._name:
            base_name, part, suffix = clef_name.partition("^")
            if suffix == "8":
                alteration = -7
            elif suffix == "15":
                alteration = -13
            else:
                raise Exception(f"bad clef alteration suffix: {suffix!r}.")
        else:
            base_name = clef_name
        return self._clef_name_to_middle_c_position[base_name] + alteration

    def _clef_name_to_staff_position_zero(self, clef_name):
        return {
            "treble": NamedPitch("B4"),
            "alto": NamedPitch("C4"),
            "varC": NamedPitch("C4"),
            "tenor": NamedPitch("A3"),
            "tenorvarC": NamedPitch("A3"),
            "bass": NamedPitch("D3"),
            "french": NamedPitch("D5"),
            "soprano": NamedPitch("G4"),
            "mezzosoprano": NamedPitch("E4"),
            "baritone": NamedPitch("F3"),
            "varbaritone": NamedPitch("F3"),
            "percussion": None,
            "tab": None,
        }[clef_name]

    def _get_format_specification(self):
        return FormatSpecification(
            self,
            repr_is_indented=False,
            storage_format_args_values=[self.name],
            storage_format_is_indented=False,
        )

    def _get_lilypond_format(self):
        return rf'\clef "{self.name}"'

    def _get_lilypond_format_bundle(self, component=None):
        bundle = LilyPondFormatBundle()
        if not self.hide:
            bundle.opening.commands.append(self._get_lilypond_format())
        return bundle

    @classmethod
    def _list_clef_names(class_):
        return list(sorted(class_._clef_name_to_middle_c_position))

    ### PUBLIC METHODS ###

    @staticmethod
    def from_pitches(pitches) -> "Clef":
        """
        Makes clef from ``pitches``.

        ..  container:: example

            >>> maker = abjad.NoteMaker()
            >>> notes = maker(range(-12, -6), [(1, 4)])
            >>> staff = abjad.Staff(notes)
            >>> pitches = abjad.iterate(staff).pitches()
            >>> abjad.Clef.from_pitches(pitches)
            Clef('bass')

            Choses between treble and bass based on minimal number of ledger lines.

        """
        diatonic_pitch_numbers = [
            pitch._get_diatonic_pitch_number() for pitch in pitches
        ]
        max_diatonic_pitch_number = max(diatonic_pitch_numbers)
        min_diatonic_pitch_number = min(diatonic_pitch_numbers)
        lowest_treble_line_pitch = NamedPitch("E4")
        lowest_treble_line_diatonic_pitch_number = (
            lowest_treble_line_pitch._get_diatonic_pitch_number()
        )
        candidate_steps_below_treble = (
            lowest_treble_line_diatonic_pitch_number - min_diatonic_pitch_number
        )
        highest_bass_line_pitch = NamedPitch("A3")
        highest_bass_line_diatonic_pitch_number = (
            highest_bass_line_pitch._get_diatonic_pitch_number()
        )
        candidate_steps_above_bass = (
            max_diatonic_pitch_number - highest_bass_line_diatonic_pitch_number
        )
        if candidate_steps_above_bass < candidate_steps_below_treble:
            return Clef("bass")
        else:
            return Clef("treble")

    ### PUBLIC PROPERTIES ###

    @property
    def context(self) -> str:
        """
        Gets (historically conventional) context.

        ..  container:: example

            >>> clef = abjad.Clef('treble')
            >>> clef.context
            'Staff'

        Override with ``abjad.attach(..., context='...')``.
        """
        return self._context

    @property
    def hide(self) -> typing.Optional[bool]:
        r"""
        Is true when clef should not appear in output (but should still
        determine effective clef).

        ..  container:: example

            >>> staff = abjad.Staff("c'4 d' e' f'")
            >>> abjad.attach(abjad.Clef('treble'), staff[0])
            >>> abjad.attach(abjad.Clef('alto', hide=True), staff[2])
            >>> abjad.show(staff) # doctest: +SKIP

            >>> string = abjad.lilypond(staff)
            >>> print(string)
            \new Staff
            {
                \clef "treble"
                c'4
                d'4
                e'4
                f'4
            }

            >>> for leaf in abjad.iterate(staff).leaves():
            ...     leaf, abjad.get.effective(leaf, abjad.Clef)
            ...
            (Note("c'4"), Clef('treble'))
            (Note("d'4"), Clef('treble'))
            (Note("e'4"), Clef('alto', hide=True))
            (Note("f'4"), Clef('alto', hide=True))

        """
        return self._hide

    @property
    def middle_c_position(self) -> int:
        """
        Gets middle C position of clef.

        ..  container:: example

            Gets staff position of middle C in treble clef:

            >>> abjad.Clef('treble').middle_c_position
            StaffPosition(-6)

        ..  container:: example

            Gets staff position of middle C in alto clef:

            >>> abjad.Clef('alto').middle_c_position
            StaffPosition(0)

        """
        return self._middle_c_position

    @property
    def name(self) -> str:
        """
        Gets name of clef.

        ..  container:: example

            Gets name treble clef:

            >>> abjad.Clef('treble').name
            'treble'

        ..  container:: example

            Gets name of alto clef:

            >>> abjad.Clef('alto').name
            'alto'

        """
        return self._name

    @property
    def persistent(self) -> bool:
        """
        Is true.

        ..  container:: example

            >>> abjad.Clef('treble').persistent
            True

        Class constant.
        """
        return self._persistent

    @property
    def redraw(self) -> bool:
        """
        Is true.

        ..  container:: example

            >>> abjad.Clef('treble').redraw
            True

        Class constant.
        """
        return self._redraw

    @property
    def tweaks(self) -> None:
        r"""
        Are not implemented on clef.

        The LilyPond ``\clef`` command refuses tweaks.

        Override the LilyPond ``Clef`` grob instead.
        """
        pass


@functools.total_ordering
class StaffPosition:
    """
    Staff position.

    ..  container:: example

        Initializes staff position at middle line of staff:

        >>> abjad.StaffPosition(0)
        StaffPosition(0)

    ..  container:: example

        Initializes staff position one space below middle line of staff:

        >>> abjad.StaffPosition(-1)
        StaffPosition(-1)

    ..  container:: example

        Initializes staff position one line below middle line of staff:

        >>> abjad.StaffPosition(-2)
        StaffPosition(-2)

    ..  container:: example

        Initializes from other staff position:

        >>> staff_position = abjad.StaffPosition(-2)
        >>> abjad.StaffPosition(staff_position)
        StaffPosition(-2)

    """

    ### CLASS VARIABLES ###

    __slots__ = ("_number",)

    ### INITIALIZER ###

    def __init__(self, number=0):
        if isinstance(number, type(self)):
            number = number.number
        assert isinstance(number, numbers.Number), repr(number)
        self._number = number

    ### SPECIAL METHODS ###

    def __eq__(self, argument):
        """
        Is true when ``argument`` is a staff position with the same number as
        this staff position.

        ..  container:: example

            >>> staff_position_1 = abjad.StaffPosition(-2)
            >>> staff_position_2 = abjad.StaffPosition(-2)
            >>> staff_position_3 = abjad.StaffPosition(0)

            >>> staff_position_1 == staff_position_1
            True
            >>> staff_position_1 == staff_position_2
            True
            >>> staff_position_1 == staff_position_3
            False

            >>> staff_position_2 == staff_position_1
            True
            >>> staff_position_2 == staff_position_2
            True
            >>> staff_position_2 == staff_position_3
            False

            >>> staff_position_3 == staff_position_1
            False
            >>> staff_position_3 == staff_position_2
            False
            >>> staff_position_3 == staff_position_3
            True

        Returns true or false.
        """
        return StorageFormatManager.compare_objects(self, argument)

    def __hash__(self):
        """
        Hashes staff position.

        Returns integer.
        """
        hash_values = StorageFormatManager(self).get_hash_values()
        try:
            result = hash(hash_values)
        except TypeError:
            raise TypeError(f"unhashable type: {self}")
        return result

    def __lt__(self, argument):
        """
        Is true when staff position is less than ``argument``.

        ..  container:: example

            >>> staff_position_1 = abjad.StaffPosition(-2)
            >>> staff_position_2 = abjad.StaffPosition(-2)
            >>> staff_position_3 = abjad.StaffPosition(0)

            >>> staff_position_1 < staff_position_1
            False
            >>> staff_position_1 < staff_position_2
            False
            >>> staff_position_1 < staff_position_3
            True

            >>> staff_position_2 < staff_position_1
            False
            >>> staff_position_2 < staff_position_2
            False
            >>> staff_position_2 < staff_position_3
            True

            >>> staff_position_3 < staff_position_1
            False
            >>> staff_position_3 < staff_position_2
            False
            >>> staff_position_3 < staff_position_3
            False

        Returns true or false.
        """
        try:
            argument = type(self)(argument)
        except Exception:
            return False
        return self.number < argument.number

    def __repr__(self) -> str:
        """
        Gets interpreter representation.
        """
        return StorageFormatManager(self).get_repr_format()

    def __str__(self):
        """
        Gets string representation of staff position.

        ..  container:: example

            >>> str(abjad.StaffPosition(-2))
            'StaffPosition(-2)'

        Returns string.
        """
        return f"{type(self).__name__}({self.number})"

    ### PRIVATE METHODS ###

    def _get_format_specification(self):
        return FormatSpecification(
            client=self,
            repr_is_indented=False,
            storage_format_is_indented=False,
            storage_format_args_values=[self.number],
            storage_format_keyword_names=[],
        )

    ### PUBLIC PROPERTIES ###

    @property
    def number(self):
        """
        Gets staff position number.

        ..  container:: example

            >>> abjad.StaffPosition(-2).number
            -2

        Returns number.
        """
        return self._number

    ### PUBLIC METHODS ###

    @staticmethod
    def from_pitch_and_clef(pitch, clef) -> "StaffPosition":
        r"""
        Changes named pitch to staff position.

        ..  container:: example

            Changes C#5 to absolute staff position:

            >>> abjad.StaffPosition.from_pitch_and_clef('C#5', "alto")
            StaffPosition(7)

            >>> abjad.StaffPosition.from_pitch_and_clef('C#5', "treble")
            StaffPosition(1)

            >>> abjad.StaffPosition.from_pitch_and_clef('C#5', "bass")
            StaffPosition(13)

        ..  container:: example

            Marks up absolute staff position of many pitches:

            >>> string = "g16 a b c' d' e' f' g' a' b' c'' d'' e'' f'' g'' a''"
            >>> staff = abjad.Staff(string)
            >>> for note in staff:
            ...     staff_position = abjad.StaffPosition.from_pitch_and_clef(
            ...         note.written_pitch,
            ...         "alto",
            ...     )
            ...     markup = abjad.Markup(staff_position.number)
            ...     abjad.attach(markup, note)
            ...
            >>> abjad.override(staff).TextScript.staff_padding = 5
            >>> abjad.show(staff) # doctest: +SKIP

            ..  docs::

                >>> string = abjad.lilypond(staff)
                >>> print(string)
                \new Staff
                \with
                {
                    \override TextScript.staff-padding = 5
                }
                {
                    g16
                    - \markup { -3 }
                    a16
                    - \markup { -2 }
                    b16
                    - \markup { -1 }
                    c'16
                    - \markup { 0 }
                    d'16
                    - \markup { 1 }
                    e'16
                    - \markup { 2 }
                    f'16
                    - \markup { 3 }
                    g'16
                    - \markup { 4 }
                    a'16
                    - \markup { 5 }
                    b'16
                    - \markup { 6 }
                    c''16
                    - \markup { 7 }
                    d''16
                    - \markup { 8 }
                    e''16
                    - \markup { 9 }
                    f''16
                    - \markup { 10 }
                    g''16
                    - \markup { 11 }
                    a''16
                    - \markup { 12 }
                }

        ..  container:: example

            Marks up bass staff position of many pitches:

            >>> string = "g,16 a, b, c d e f g a b c' d' e' f' g' a'"
            >>> staff = abjad.Staff(string)
            >>> for note in staff:
            ...     staff_position = abjad.StaffPosition.from_pitch_and_clef(
            ...         note.written_pitch,
            ...         "bass",
            ...     )
            ...     markup = abjad.Markup(staff_position.number)
            ...     abjad.attach(markup, note)
            ...
            >>> abjad.attach(abjad.Clef("bass"), staff[0])
            >>> abjad.override(staff).TextScript.staff_padding = 5
            >>> abjad.show(staff) # doctest: +SKIP

            ..  docs::

                >>> string = abjad.lilypond(staff)
                >>> print(string)
                \new Staff
                \with
                {
                    \override TextScript.staff-padding = 5
                }
                {
                    \clef "bass"
                    g,16
                    - \markup { -4 }
                    a,16
                    - \markup { -3 }
                    b,16
                    - \markup { -2 }
                    c16
                    - \markup { -1 }
                    d16
                    - \markup { 0 }
                    e16
                    - \markup { 1 }
                    f16
                    - \markup { 2 }
                    g16
                    - \markup { 3 }
                    a16
                    - \markup { 4 }
                    b16
                    - \markup { 5 }
                    c'16
                    - \markup { 6 }
                    d'16
                    - \markup { 7 }
                    e'16
                    - \markup { 8 }
                    f'16
                    - \markup { 9 }
                    g'16
                    - \markup { 10 }
                    a'16
                    - \markup { 11 }
                }

        """
        pitch = NamedPitch(pitch)
        clef = Clef(clef)
        staff_position_number = pitch._get_diatonic_pitch_number()
        staff_position_number += clef.middle_c_position.number
        staff_position = StaffPosition(staff_position_number)
        return staff_position

    def to_pitch(self, clef):
        """
        Makes named pitch from staff position and ``clef``.

        ..  container:: example

            Treble clef:

            >>> for n in range(-6, 6):
            ...     staff_position = abjad.StaffPosition(n)
            ...     pitch = staff_position.to_pitch("treble")
            ...     message = f"{staff_position!s}    {pitch}"
            ...     print(message)
            ...
            StaffPosition(-6)    c'
            StaffPosition(-5)    d'
            StaffPosition(-4)    e'
            StaffPosition(-3)    f'
            StaffPosition(-2)    g'
            StaffPosition(-1)    a'
            StaffPosition(0)    b'
            StaffPosition(1)    c''
            StaffPosition(2)    d''
            StaffPosition(3)    e''
            StaffPosition(4)    f''
            StaffPosition(5)    g''

        ..  container:: example

            Bass clef:

            >>> for n in range(-6, 6):
            ...     staff_position = abjad.StaffPosition(n)
            ...     pitch = staff_position.to_pitch("bass")
            ...     message = f"{staff_position!s}    {pitch}"
            ...     print(message)
            ...
            StaffPosition(-6)    e,
            StaffPosition(-5)    f,
            StaffPosition(-4)    g,
            StaffPosition(-3)    a,
            StaffPosition(-2)    b,
            StaffPosition(-1)    c
            StaffPosition(0)    d
            StaffPosition(1)    e
            StaffPosition(2)    f
            StaffPosition(3)    g
            StaffPosition(4)    a
            StaffPosition(5)    b

        ..  container:: example

            Alto clef:

            >>> for n in range(-6, 6):
            ...     staff_position = abjad.StaffPosition(n)
            ...     pitch = staff_position.to_pitch("alto")
            ...     message = f"{staff_position!s}    {pitch}"
            ...     print(message)
            ...
            StaffPosition(-6)    d
            StaffPosition(-5)    e
            StaffPosition(-4)    f
            StaffPosition(-3)    g
            StaffPosition(-2)    a
            StaffPosition(-1)    b
            StaffPosition(0)    c'
            StaffPosition(1)    d'
            StaffPosition(2)    e'
            StaffPosition(3)    f'
            StaffPosition(4)    g'
            StaffPosition(5)    a'

        ..  container:: example

            Percussion clef:

            >>> for n in range(-6, 6):
            ...     staff_position = abjad.StaffPosition(n)
            ...     pitch = staff_position.to_pitch("percussion")
            ...     message = f"{staff_position!s}    {pitch}"
            ...     print(message)
            ...
            StaffPosition(-6)    d
            StaffPosition(-5)    e
            StaffPosition(-4)    f
            StaffPosition(-3)    g
            StaffPosition(-2)    a
            StaffPosition(-1)    b
            StaffPosition(0)    c'
            StaffPosition(1)    d'
            StaffPosition(2)    e'
            StaffPosition(3)    f'
            StaffPosition(4)    g'
            StaffPosition(5)    a'

        Returns new named pitch.
        """
        offset_staff_position_number = self.number
        clef = Clef(clef)
        offset_staff_position_number -= clef.middle_c_position.number
        offset_staff_position = StaffPosition(offset_staff_position_number)
        octave_number = offset_staff_position.number // 7 + 4
        diatonic_pc_number = offset_staff_position.number % 7
        pitch_class_number = pitch__lib._diatonic_pc_number_to_pitch_class_number[
            diatonic_pc_number
        ]
        pitch_number = 12 * (octave_number - 4)
        pitch_number += pitch_class_number
        named_pitch = NamedPitch(pitch_number)
        return named_pitch



import collections

from .storage import FormatSpecification, StorageFormatManager


class OrderedDict(collections.abc.MutableMapping):
    r"""
    Ordered dictionary.

    ..  container:: example

        Initializes from list of pairs:

        >>> string = r"\markup \italic Allegretto"
        >>> markup = abjad.Markup(string, literal=True)
        >>> dictionary = abjad.OrderedDict([
        ...     ('color', 'red'),
        ...     ('directive', markup),
        ...     ])

        >>> string = abjad.storage(dictionary)
        >>> print(string)
        abjad.OrderedDict(
            [
                ('color', 'red'),
                (
                    'directive',
                    abjad.Markup(
                        contents=['\\markup \\italic Allegretto'],
                        literal=True,
                        ),
                    ),
                ]
            )

    ..  container:: example

        Initializes from built-in dictionary:

        >>> string = r"\markup \italic Allegretto"
        >>> markup = abjad.Markup(string, literal=True)
        >>> dictionary = {
        ...     'color': 'red',
        ...     'directive': markup,
        ...     }
        >>> dictionary = abjad.OrderedDict(
        ...     dictionary
        ...     )

        >>> string = abjad.storage(dictionary)
        >>> print(string)
        abjad.OrderedDict(
            [
                ('color', 'red'),
                (
                    'directive',
                    abjad.Markup(
                        contents=['\\markup \\italic Allegretto'],
                        literal=True,
                        ),
                    ),
                ]
            )

    ..  container:: example

        Initializes from other ordered dictionary:

        >>> string = r"\markup \italic Allegretto"
        >>> markup = abjad.Markup(string, literal=True)
        >>> dictionary_1 = abjad.OrderedDict([
        ...     ('color', 'red'),
        ...     ('directive', markup),
        ...     ])
        >>> dictionary_2 = abjad.OrderedDict(
        ...     dictionary_1
        ...     )

        >>> string = abjad.storage(dictionary_2)
        >>> print(string)
        abjad.OrderedDict(
            [
                ('color', 'red'),
                (
                    'directive',
                    abjad.Markup(
                        contents=['\\markup \\italic Allegretto'],
                        literal=True,
                        ),
                    ),
                ]
            )

    """

    ### CLASS VARIABLES ###

    __slots__ = ("_collection",)

    ### INITIALIZER ###

    def __init__(self, items=None):
        if isinstance(items, dict):
            items = sorted(items.items())
        elif isinstance(items, collections.abc.Mapping):
            items = list(items.items())
        items = items or []
        the_items = []
        for item in items:
            assert len(item) == 2, repr(item)
            key = item[0]
            value = item[1]
            the_item = (key, value)
            the_items.append(the_item)
        self._collection = collections.OrderedDict(the_items)

    ### SPECIAL METHODS ###

    def __cmp__(self, argument):
        """
        Aliases OrderedDict.__cmp__().

        Returns true or false.
        """
        assert isinstance(argument, type(self))
        ordered_dictionary = argument._collection
        return self._collection.__cmp__(ordered_dictionary)

    def __contains__(self, key):
        """
        Aliases OrderedDict.__contains__().

        Returns true or false.
        """
        return key in self._collection

    def __delitem__(self, i):
        """
        Aliases OrderedDict.__delitem__().

        Returns none.
        """
        del self._collection[i]

    def __eq__(self, argument):
        """
        Is true when ``argument`` is an ordered dict with items that
        compare equal to those of this ordered dict.

        Returns true or false.
        """
        if issubclass(type(argument), type(self)):
            return self._collection == argument._collection
        elif isinstance(argument, type(self._collection)):
            return self._collection == argument
        return False

    def __ge__(self, argument):
        """
        Is true when typed ordered dictionary is greater than or equal
        to ``argument``.

        Returns true or false.
        """
        argument = type(self)(argument)
        return self._collection.__ge__(argument._collection)

    def __getitem__(self, argument):
        """
        Gets item or slice identified by ``argument``.

        Returns item or slice.
        """
        return self._collection.__getitem__(argument)

    def __gt__(self, argument):
        """
        Is true when typed ordered dictionary is greater than ``argument``.

        Returns true or false.
        """
        argument = type(self)(argument)
        return self._collection.__gt__(argument._collection)

    def __hash__(self):
        """
        Hashes typed collection.

        Redefined in tandem with __eq__.
        """
        return object.__hash__(self)

    def __iter__(self):
        """
        Iterates typed collection.

        Returns generator.
        """
        return self._collection.__iter__()

    def __len__(self):
        """
        Gets length of typed collection.

        Returns nonnegative integer.
        """
        return len(self._collection)

    def __le__(self, argument):
        """
        Is true when typed ordered dictionary is less than or equal
        to ``argument``.

        Returns true or false.
        """
        argument = type(self)(argument)
        return self._collection.__le__(argument._collection)

    def __lt__(self, argument):
        """
        Is true when typed ordered dictionary is less than ``argument``.

        Returns true or false.
        """
        argument = type(self)(argument)
        return self._collection.__lt__(argument._collection)

    def __reversed__(self):
        """
        Aliases OrderedDict.__reversed__().

        Returns generatos.
        """
        return self._collection.__reversed__()

    def __repr__(self) -> str:
        """
        Gets interpreter representation.
        """
        return StorageFormatManager(self).get_repr_format()

    def __setitem__(self, i, argument):
        """
        Changes items in ``argument`` to items and sets.

        Returns none.
        """
        self._collection[i] = argument

    ### PRIVATE METHODS ###

    def _get_format_specification(self):
        names = list(StorageFormatManager(self).signature_keyword_names)
        if "items" in names:
            names.remove("items")
        values = [list(self._collection.items())]
        return FormatSpecification(
            self,
            repr_is_indented=False,
            storage_format_args_values=values,
            storage_format_keyword_names=names,
        )

    ### PUBLIC METHODS ###

    def clear(self):
        """
        Clears typed ordered dictionary.

        Returns none.
        """
        self._collection.clear()

    def copy(self):
        """
        Copies typed ordered dictionary.

        Returns new typed ordered dictionary.
        """
        ordered_dictionary = self._collection.copy()
        items = list(ordered_dictionary.items())
        return type(self)(items=items)

    def get(self, i, default=None):
        """
        Aliases OrderedDict.get().

        Returns item or raises key error.
        """
        return self._collection.get(i, default)

    def has_key(self, key):
        """
        Aliases OrderdDict.has_key().

        Returns true or false.
        """
        return key in self._collection

    def items(self):
        """
        Aliases OrderedDict.items().

        Returns generator.
        """
        return iter(self._collection.items())

    def keys(self):
        """
        Aliases OrderedDict.keys().

        Returns generator.
        """
        return iter(self._collection.keys())

    def pop(self, key, default=None):
        """
        Aliases OrderedDict.pop().

        Returns items.
        """
        return self._collection.pop(key, default)

    def popitem(self):
        """
        Aliases OrderedDict.popitem().

        Returns generator.
        """
        return self._collection.popitem()

    def setdefault(self, key, default=None):
        """
        Aliases OrderedDict.setdefault().

        Returns items.
        """
        return self._collection.setdefault(key, default)

    def sort(self, recurse=False) -> None:
        """
        Sorts ordered dictionary (in place).

        ..  container:: example

            >>> dictionary = abjad.OrderedDict()
            >>> dictionary['flavor'] = 'cherry'
            >>> dictionary['colors'] = abjad.OrderedDict()
            >>> dictionary['colors']['red'] = 3
            >>> dictionary['colors']['green'] = 2
            >>> dictionary['colors']['blue'] = 1
            >>> string = abjad.storage(dictionary)
            >>> print(string)
            abjad.OrderedDict(
                [
                    ('flavor', 'cherry'),
                    (
                        'colors',
                        abjad.OrderedDict(
                            [
                                ('red', 3),
                                ('green', 2),
                                ('blue', 1),
                                ]
                            ),
                        ),
                    ]
                )

            >>> dictionary.sort()
            >>> string = abjad.storage(dictionary)
            >>> print(string)
            abjad.OrderedDict(
                [
                    (
                        'colors',
                        abjad.OrderedDict(
                            [
                                ('red', 3),
                                ('green', 2),
                                ('blue', 1),
                                ]
                            ),
                        ),
                    ('flavor', 'cherry'),
                    ]
                )

            >>> dictionary.sort(recurse=True)
            >>> string = abjad.storage(dictionary)
            >>> print(string)
            abjad.OrderedDict(
                [
                    (
                        'colors',
                        abjad.OrderedDict(
                            [
                                ('blue', 1),
                                ('green', 2),
                                ('red', 3),
                                ]
                            ),
                        ),
                    ('flavor', 'cherry'),
                    ]
                )

        """
        items = list(self.items())
        items.sort()
        self.clear()
        for key, value in items:
            if recurse is True and isinstance(value, OrderedDict):
                value.sort(recurse=True)
            self[key] = value

    def update(self, *arguments, **keywords):
        """
        Aliases OrderedDict.update().

        Returns none.
        """
        return self._collection.update(*arguments, **keywords)

    def values(self):
        """
        Aliases OrderedDict.values().

        Returns generator.
        """
        return iter(self._collection.values())



import copy

from . import deprecated, enums, get, overrides, score, select
from . import timespan as _timespan
from .attach import attach
from .duration import Duration
from .indicators.Clef import Clef
from .indicators.StaffChange import StaffChange
from .iterate import Iteration
from .lilypond import lilypond
from .lilypondfile import Block, LilyPondFile
from .makers import NoteMaker
from .markups import Markup, Postscript
from .metricmodulation import MetricModulation
from .new import new
from .ordereddict import OrderedDict
from .pitch.PitchRange import PitchRange
from .pitch.pitches import NamedPitch
from .pitch.segments import PitchClassSegment, PitchSegment
from .pitch.sets import PitchClassSet, PitchSet
from .spanners import glissando


def _illustrate_component(component):
    block = Block(name="score")
    block.items.append(component)
    lilypond_file = LilyPondFile(items=[block])
    return lilypond_file


def _illustrate_markup(markup):
    lilypond_file = LilyPondFile()
    markup = new(markup, direction=None)
    lilypond_file.items.append(markup)
    return lilypond_file


def _illustrate_markup_maker(argument, **keywords):
    markup = argument._make_markup(**keywords)
    return _illustrate_markup(markup)


def _illustrate_postscript(postscript):
    markup = Markup.postscript(postscript)
    return _illustrate_markup(markup)


def _illustrate_metric_modulation(metric_modulation):
    lilypond_file = LilyPondFile()
    markup = metric_modulation._get_markup()
    lilypond_file.items.append(markup)
    return lilypond_file


def _illustrate_pitch_class_set(set_):
    chord = score.Chord(set_, Duration(1))
    voice = score.Voice([chord])
    staff = score.Staff([voice])
    score_ = score.Score([staff])
    lilypond_file = LilyPondFile(items=[score_])
    return lilypond_file


def _illustrate_pitch_range(pitch_range):
    start_pitch_clef = Clef.from_pitches([pitch_range.start_pitch])
    stop_pitch_clef = Clef.from_pitches([pitch_range.stop_pitch])
    start_note = score.Note(pitch_range.start_pitch, 1)
    stop_note = score.Note(pitch_range.stop_pitch, 1)
    if start_pitch_clef == stop_pitch_clef:
        if start_pitch_clef == Clef("bass"):
            bass_staff = score.Staff()
            attach(Clef("bass"), bass_staff)
            bass_staff.extend([start_note, stop_note])
            bass_leaves = select.Selection(bass_staff).leaves()
            glissando(bass_leaves)
            score_ = score.Score([bass_staff])
        else:
            treble_staff = score.Staff()
            attach(Clef("treble"), treble_staff)
            treble_staff.extend([start_note, stop_note])
            treble_leaves = select.Selection(treble_staff).leaves()
            glissando(treble_leaves)
            score_ = score.Score([treble_staff])
    else:
        result = make_piano_score()
        score_, treble_staff, bass_staff = result
        bass_staff.extend([start_note, stop_note])
        treble_staff.extend("s1 s1")
        bass_leaves = select.Selection(bass_staff).leaves()
        glissando(bass_leaves)
        attach(StaffChange("Treble_Staff"), bass_staff[1])
        attach(Clef("treble"), treble_staff[0])
        attach(Clef("bass"), bass_staff[0])
    for leaf in Iteration(score_).leaves():
        leaf.multiplier = (1, 4)
    overrides.override(score_).BarLine.stencil = False
    overrides.override(score_).SpanBar.stencil = False
    overrides.override(score_).Glissando.thickness = 2
    overrides.override(score_).TimeSignature.stencil = False
    lilypond_file = LilyPondFile(items=[score_])
    return lilypond_file


def _illustrate_pitch_segment(segment):
    named_pitches = [NamedPitch(x) for x in segment]
    maker = NoteMaker()
    notes = maker(named_pitches, [1])
    result = make_piano_score(leaves=notes, sketch=True)
    score_, treble_staff, bass_staff = result
    for leaf in Iteration(score_).leaves():
        leaf.multiplier = (1, 8)
    overrides.override(score_).Rest.transparent = True
    lilypond_file = LilyPondFile(items=[score_])
    return lilypond_file


def _illustrate_pitch_set(set_):
    upper, lower = [], []
    for pitch in set_:
        if pitch < 0:
            lower.append(pitch)
        else:
            upper.append(pitch)
    if upper:
        upper = score.Chord(upper, Duration(1))
    else:
        upper = score.Skip((1, 1))
    if lower:
        lower = score.Chord(lower, Duration(1))
    else:
        lower = score.Skip((1, 1))
    upper_voice = score.Voice([upper])
    upper_staff = score.Staff([upper_voice])
    lower_voice = score.Voice([lower])
    lower_staff = score.Staff([lower_voice])
    staff_group = score.StaffGroup(
        [upper_staff, lower_staff], lilypond_type="PianoStaff"
    )
    score_ = score.Score([staff_group])
    lilypond_file = LilyPondFile(items=[score_])
    return lilypond_file


def _illustrate_pitch_class_segment(
    segment, markup_direction=enums.Up, figure_name=None
):
    notes = []
    for item in segment:
        note = score.Note(item, Duration(1, 8))
        notes.append(note)
    markup = None
    if segment._equivalence_markup:
        markup = segment._equivalence_markup
    if isinstance(figure_name, str):
        figure_name = Markup(figure_name)
    if figure_name is not None:
        markup = figure_name
    if markup is not None:
        direction = markup_direction
        markup = new(markup, direction=direction)
        attach(markup, notes[0])
    voice = score.Voice(notes)
    staff = score.Staff([voice])
    score_ = score.Score([staff])
    preamble = r"""\layout {
    \accidentalStyle forget
    indent = 0
    \context {
        \Score
        \override BarLine.transparent = ##t
        \override BarNumber.stencil = ##f
        \override Beam.stencil = ##f
        \override Flag.stencil = ##f
        \override Stem.stencil = ##f
        \override TimeSignature.stencil = ##f
        proportionalNotationDuration = #(ly:make-moment 1 12)
    }
}

\paper {
    markup-system-spacing.padding = 8
    system-system-spacing.padding = 10
    top-markup-spacing.padding = 4
}"""
    deprecated.add_final_bar_line(score_)
    string = r"\override Score.BarLine.transparent = ##f"
    command = overrides.LilyPondLiteral(string, "after")
    last_leaf = select.Selection(score_).leaves()[-1]
    attach(command, last_leaf)
    lilypond_file = LilyPondFile(items=[preamble, score_])
    return lilypond_file


def _illustrate_timespan(timespan):
    timespans = _timespan.TimespanList([timespan])
    return _illustrate_markup_maker(timespans)


_class_to_method = OrderedDict(
    [
        (score.Component, _illustrate_component),
        (Markup, _illustrate_markup),
        (MetricModulation, _illustrate_metric_modulation),
        (_timespan.OffsetCounter, _illustrate_markup_maker),
        (Postscript, _illustrate_postscript),
        (PitchRange, _illustrate_pitch_range),
        (PitchClassSet, _illustrate_pitch_class_set),
        (PitchSegment, _illustrate_pitch_segment),
        (PitchClassSegment, _illustrate_pitch_class_segment),
        (PitchSet, _illustrate_pitch_set),
        (_timespan.Timespan, _illustrate_timespan),
        (_timespan.TimespanList, _illustrate_markup_maker),
    ]
)


### PUBLIC FUNCTIONS ###


def attach_markup_struts(lilypond_file):
    """
    LilyPond workaround.

    LilyPond's multisystem cropping currently removes intersystem whitespace.

    These transparent markup struts force LilyPond's cropping to preserve whitespace.
    """
    rhythmic_staff = lilypond_file[score.Score][-1]
    first_leaf = get.leaf(rhythmic_staff, 0)
    markup = Markup(r"\markup I", direction=enums.Up, literal=True)
    attach(markup, first_leaf)
    overrides.tweak(markup).staff_padding = 11
    overrides.tweak(markup).transparent = "##t"
    duration = get.duration(rhythmic_staff)
    if Duration(6, 4) < duration:
        last_leaf = get.leaf(rhythmic_staff, -1)
        markup = Markup(r"\markup I", direction=enums.Up, literal=True)
        attach(markup, last_leaf)
        overrides.tweak(markup).staff_padding = 18
        overrides.tweak(markup).transparent = "##t"


def illustrate(item, **keywords):
    """
    Illustrates ``item``.
    """
    method = None
    for key in _class_to_method:
        if isinstance(item, key):
            method = _class_to_method[key]
            break
    if method is None:
        raise Exception(f"can not illustrate objects of type {type(item)}.")
    return method(item, **keywords)


def make_piano_score(leaves=None, lowest_treble_pitch="B3", sketch=False):
    """
    Makes piano score from ``leaves``.
    """
    leaves = leaves or []
    lowest_treble_pitch = NamedPitch(lowest_treble_pitch)
    treble_staff = score.Staff(name="Treble_Staff")
    bass_staff = score.Staff(name="Bass_Staff")
    staff_group = score.StaffGroup(
        [treble_staff, bass_staff], lilypond_type="PianoStaff"
    )
    score_ = score.Score()
    score_.append(staff_group)
    for leaf in leaves:
        treble_pitches, bass_pitches = [], []
        selection = select.Selection(leaf)
        pitch_set = PitchSet.from_selection(selection)
        for pitch in pitch_set:
            if pitch < lowest_treble_pitch:
                bass_pitches.append(pitch)
            else:
                treble_pitches.append(pitch)
        written_duration = leaf.written_duration
        if not treble_pitches:
            treble_leaf = score.Rest(written_duration)
        elif len(treble_pitches) == 1:
            treble_leaf = score.Note(treble_pitches[0], written_duration)
        else:
            treble_leaf = score.Chord(treble_pitches, written_duration)
        treble_staff.append(treble_leaf)
        if not bass_pitches:
            bass_leaf = score.Rest(written_duration)
        elif len(bass_pitches) == 1:
            bass_leaf = score.Note(bass_pitches[0], written_duration)
        else:
            bass_leaf = score.Chord(bass_pitches, written_duration)
        bass_staff.append(bass_leaf)
    if 0 < len(treble_staff):
        attach(Clef("treble"), treble_staff[0])
    if 0 < len(bass_staff):
        attach(Clef("bass"), bass_staff[0])
    if sketch:
        overrides.override(score_).TimeSignature.stencil = False
        overrides.override(score_).BarNumber.transparent = True
        overrides.override(score_).BarLine.stencil = False
        overrides.override(score_).SpanBar.stencil = False
    return score_, treble_staff, bass_staff


def selection_to_score_markup_string(selection):
    """
    Changes ``selection`` to score markup string.
    """
    selection = copy.deepcopy(selection)
    staff = score.Staff(selection)
    staff.lilypond_type = "RhythmicStaff"
    staff.remove_commands.append("Time_signature_engraver")
    staff.remove_commands.append("Staff_symbol_engraver")
    overrides.override(staff).Stem.direction = enums.Up
    overrides.override(staff).Stem.length = 5
    overrides.override(staff).TupletBracket.bracket_visibility = True
    overrides.override(staff).TupletBracket.direction = enums.Up
    overrides.override(staff).TupletBracket.minimum_length = 4
    overrides.override(staff).TupletBracket.padding = 1.25
    overrides.override(staff).TupletBracket.shorten_pair = "#'(-1 . -1.5)"
    scheme = "#ly:spanner::set-spacing-rods"
    overrides.override(staff).TupletBracket.springs_and_rods = scheme
    overrides.override(staff).TupletNumber.font_size = 0
    scheme = "#tuplet-number::calc-fraction-text"
    overrides.override(staff).TupletNumber.text = scheme
    overrides.setting(staff).tupletFullLength = True
    layout_block = Block(name="layout")
    layout_block.indent = 0
    layout_block.ragged_right = "##t"
    score_ = score.Score([staff])
    overrides.override(score_).SpacingSpanner.spacing_increment = 0.5
    overrides.setting(score_).proportionalNotationDuration = False
    indent = 4 * " "
    strings = [r"\score", indent + "{"]
    strings.extend([2 * indent + _ for _ in lilypond(score_).split("\n")])
    strings.extend([2 * indent + _ for _ in lilypond(layout_block).split("\n")])
    strings.append(indent + "}")
    string = "\n".join(strings)
    return string
